#! /usr/bin/env python3

import argparse
import copy
import datetime
import glob
import json
import operator
import os
import pickle
import re
import readline
import sys
from collections import defaultdict
from functools import reduce

from snakemake.common.configfile import load_configfile

from MONSDA.Logger import *

from . import _version

__version__ = _version.get_versions()["version"]

try:
    pythonversion = f"python{str(sys.version_info.major)}.{str(sys.version_info.minor)}"
    installpath = os.path.dirname(__file__).replace(
        os.sep.join(["lib", pythonversion, "site-packages", "MONSDA"]), "share"
    )
except:
    installpath = os.path.cwd()

configpath = os.path.join(installpath, "MONSDA", "configs")
current_path = os.getcwd()
dir_path = os.path.dirname(os.path.realpath(__file__))
os.chdir(dir_path)

template = load_configfile(os.sep.join([configpath, "template_base_commented.json"]))
none_workflow_keys = ["WORKFLOWS", "BINS", "MAXTHREADS", "SETTINGS", "VERSION"]
comparable_workflows = ["DE", "DEU", "DAS", "DTU", "PEAKS"]
quantifying_workflows = ["COUNTING"]
IP_workflows = ["PEAKS"]
index_prefix_workflows = ["MAPPING"]


parser = argparse.ArgumentParser(
    description="Helper to create or manipulate configuration file or project folder used for workflow processing with MONSDA"
)
parser.add_argument(
    "-t",
    "--test",
    action="store_true",
    default=False,
    help="running in test-mode for showing interim results to copy",
)

parser.add_argument(
    "-s",
    "--session",
    type=str,
    default=False,
    help="load unfinished config session to continue",
)

parser.add_argument(
    "-c",
    "--config",
    type=str,
    default=False,
    help="takes configuration file to modify",
)

args = parser.parse_args()


class NestedDefaultDict(defaultdict):
    def __init__(self, *args, **kwargs):
        super(NestedDefaultDict, self).__init__(NestedDefaultDict, *args, **kwargs)

    def __reduce__(self):
        return (type(self), (), None, None, iter(self.items()))


class PROJECT:
    def __init__(self):
        self.mode = ""
        self.name = ""
        self.unfinished_config = "unfinished_config.pkl"
        self.current_func = ""
        self.current_func_arg = None
        self.current_wf = []
        self.finished_maplists = []
        self.finished_settings = []
        self.finished_set_keys = []
        self.finished_set_maplists = []
        self.subname = ""
        self.path = ""
        self.cores = 1
        self.baseDict = NestedDefaultDict()
        self.commentsDict = NestedDefaultDict()
        self.conditionsDict = NestedDefaultDict()
        self.samplesDict = NestedDefaultDict()
        self.workflowsDict = NestedDefaultDict()
        self.settingsDict = NestedDefaultDict()
        self.settingsList = []


class GUIDE:
    def __init__(self):
        self.answer = ""
        self.toclear = 0
        self.testing = False

    def clear(self, number=None):
        if not number:
            number = self.toclear
        CURSOR_UP_ONE = "\x1b[1A"
        ERASE_LINE = "\x1b[2K"
        for i in range(number):
            print(CURSOR_UP_ONE + ERASE_LINE + CURSOR_UP_ONE)
        self.toclear = 0

    def display(
        self, options=None, question=None, proof=None, spec=None, whitespace=False
    ):
        if options:
            space = 0
            for k, v in options.items():
                if len(str(k)) >= space:
                    space = len(str(k))
            for k, v in options.items():
                prYellow(f"   {k}{' '*(space+2-len(str(k)))}>  {v}")
                guide.toclear += 1
        if question:
            print("\n" + bold_color.BOLD + question + bold_color.END)
            guide.toclear += 1
        self.proof_input(whitespace, proof, spec)

    def proof_input(self, whitespace, proof=None, spec=None):
        unallowed_characters = []
        while True:
            if spec:
                a = rlinput(">>> ", spec)
            else:
                if whitespace:
                    a = input(">>> ").strip()
                else:
                    a = input(">>> ").strip().replace(" ", "")
            if a == "exit":
                if project.current_func:
                    pickle_unfinished(project.current_func)
                    file = os.path.join(
                        current_path,
                        "unfinished_config.pkl",
                    )
                    print(
                        f"\n\nYour entries has been saved. Continue your Session with\n"
                    )
                    prGreen(f"   monsda_configure -s {file}\n\n")
                exit()

            if any(x in unallowed_characters for x in a):
                safe = self.toclear
                self.clear(2)
                self.toclear = safe
                prRed("You used forbidden letters, try again")
                continue

            if proof == None:
                self.answer = a
                break
            elif proof == "only_numbers":
                try:
                    [float(x) for x in a.split(",")]
                    self.answer = a
                    break
                except:
                    self.clear(2)
                    prRed("please enter integer or float")
                    continue
            elif "end_exist_" in proof:
                if not a:
                    self.answer = a
                    break
                ending = proof.replace("end_exist_", "") + "$"
                if not re.findall(ending, a):
                    self.clear(2)
                    prRed(f"Nope, file has to end with '{ending.replace('$','')}'")
                    continue
                elif not os.path.isfile(a):
                    self.clear(2)
                    prRed(f"Nope, '{a}' doesn't exist")
                    continue
                else:
                    self.answer = a
                    break
            else:
                if any(x not in proof for x in a.split(",")) and not any(
                    "-" in x for x in a.split(",")
                ):
                    safe = self.toclear
                    self.clear(2)
                    self.toclear = safe
                    prRed(f"available are: {proof}")
                    continue
                else:
                    self.answer = a
                break


class bold_color:
    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


## PRINTERS
def prRed(skk):
    print("\033[91m{}\033[00m".format(skk))


def prGreen(skk):
    print("\033[92m{}\033[00m".format(skk))


def prYellow(skk):
    print("\033[93m{}\033[00m".format(skk))


def prLightPurple(skk):
    print("\033[94m{}\033[00m".format(skk))


def prPurple(skk):
    print("\033[95m{}\033[00m".format(skk))


def prCyan(skk):
    print("\033[96m{}\033[00m".format(skk))


def prLightGray(skk):
    print("\033[97m{}\033[00m".format(skk))


def print_dict(dict, indent=6, col=None, gap=""):
    if col == "Cyan":
        for line in json.dumps(dict, indent=indent).split("\n"):
            guide.toclear += 1
            prCyan(gap + line)
    else:
        for line in json.dumps(dict, indent=indent).split("\n"):
            guide.toclear += 1
            print(gap + line)


## DICT MANIPULATORS
def get_by_path(root, items):
    """Access a nested object in root by item sequence."""
    return reduce(operator.getitem, items, root)


def set_by_path(root, items, value):
    """Set a value in a nested object in root by item sequence."""
    if isinstance(get_by_path(root, items[:-1])[items[-1]], list):
        gp = get_by_path(root, items[:-1])[items[-1]]
        gp.append(value)
    else:
        get_by_path(root, items[:-1])[items[-1]] = value


def del_by_path(root, items):
    """Delete a key-value in a nested object in root by item sequence."""
    if isinstance(get_by_path(root, items[:-1])[items[-1]], str):
        get_by_path(root, items[:-1])[items[-1]] = ""
    else:
        del get_by_path(root, items[:-1])[items[-1]]


def setInDict(dataDict, maplist, value):
    first, rest = maplist[0], maplist[1:]
    if rest:
        try:
            if not isinstance(dataDict[first], dict):
                dataDict[first] = {}
        except KeyError:
            dataDict[first] = {}
        setInDict(dataDict[first], rest, value)
    else:
        try:
            if isinstance(dataDict[first], list):
                dataDict[first].append(value)
            else:
                dataDict[first] = value
        except:
            dataDict[first] = value


def get_conditions_from_dict(root, keylist=[]):
    for k, v in root.items():
        keylist.append(
            k,
        )
        if not v:
            yield ":".join(keylist)
        else:
            yield from get_conditions_from_dict(v, keylist)
        keylist.pop()


def getPathesFromDict(d, value=None):
    def yield_func(d):
        q = [(d, [])]
        while q:
            n, p = q.pop(0)
            yield p
            if isinstance(n, dict):
                for k, v in n.items():
                    q.append((v, p + [k]))
            elif isinstance(n, list):
                for i, v in enumerate(n):
                    q.append((v, p + [i]))

    ret = [p for p in yield_func(d)]
    if value:
        sel = []
        for path in ret:
            if value in path:
                sel.append(path)
        return sel
    return ret


def decouple(d):
    string = json.dumps(d)
    return json.loads(string)


def print_dict_pointer(dict, path, copy, indent=6):
    text = json.dumps(dict, indent=indent)
    route = ["step"] + path.copy()
    out = f"\n{'='*60}\n\n"
    stepper = 1
    for line in text.split("\n"):
        level = int(((len(line) - len(line.lstrip(" "))) - indent) / indent)
        key = (
            line.replace('"', "")
            .replace("{", "")
            .replace("}", "")
            .replace(":", "")
            .replace(",", "")
            .strip()
        )
        if level + 1 >= len(route):
            out += line + "\n"
        elif not key:
            out += line + "\n"
        elif route[level + 1] == key and route[level] == "step":
            route[stepper] = "step"
            stepper += 1
            if len(route) == level + 2:
                if route[level - 1] == "step":
                    if copy and copy != [""]:
                        out += f"{line}    <=(add sub conditions here)\n"
                        option = f"enter ID's on condition level comma separated \n\nor copy {copy} with 'cp'"
                        guide.toclear += 2
                    else:
                        out += f"{line}    <=(add sub conditions here)\n"
                        option = "enter ID's on conditions comma separated "
                else:
                    out += f"{line}    <=(add sub conditions here)\n"
                    option = "enter ID's on conditions comma separated "
            else:
                out += line + "\n"
        else:
            out += line + "\n"
    out += f"\n{'='*60}\n\n"
    return out, option


def rec_tree_builder(subtree, leafes, path=[], tree=None, add_mode=False):
    if tree == None:
        tree = subtree
    if not leafes[0]:
        path.pop()
        return
    for leaf in leafes:
        if not leaf in subtree.keys():
            subtree[leaf] = NestedDefaultDict()
    copy = []
    for k, v in subtree.items():
        if str(v) == "{}":
            continue
        path.append(k)
        text, opt = print_dict_pointer(tree, path, copy)
        for line in text.split("\n"):
            if "<=(add sub conditions here)" in line:
                prYellow("  " + line)
            else:
                prRed("  " + line)
        guide.toclear += len(text.split("\n")) + 2
        guide.display(question=opt)
        guide.clear()
        if guide.answer != "":
            guide.answer = guide.answer.rstrip(",")
            leafes = [x for x in guide.answer.split(",")]
        elif guide.answer == "" and isinstance(v, dict) and v != {}:
            leafes = list(v.keys())
        else:
            leafes = [""]
        if leafes == ["cp"]:
            leafes = copy
        rec_tree_builder(subtree[k], leafes, path, tree, add_mode)
        copy = leafes
        leafes = [""]
    if len(path) > 0:
        path.pop()
    return


def depth(d):
    if isinstance(d, dict):
        return 1 + (max(map(depth, d.values())) if d else 0)
    return 0


def select_id_to_set(cdict, i, indent=6):
    text = json.dumps(cdict, indent=indent)
    project.settingsList = []
    d = depth(cdict)
    path = []
    reminder = ""
    counter = 0
    prRed(f"\n{'='*60}\n")
    for line in text.split("\n"):
        level = int(((len(line) - len(line.lstrip(" "))) - indent) / indent)
        key = (
            line.replace('"', "")
            .replace("{", "")
            .replace("}", "")
            .replace(":", "")
            .replace(",", "")
            .strip()
        )
        if key:
            if len(path) > level:
                path = path[: -(len(path) - level)]
            path.append(key)
        if level == i and ":" in line:
            if reminder != key:
                counter += 1
                reminder = key
                project.settingsList.append([])
        elif level < i and "{}" in line:
            counter += 1
            project.settingsList.append([])
        if "{}" in line:
            if "," in line:
                out = f"{line}{' '*(14-len(key) + indent*(d-2)-indent*level)} <={' '*((counter+1)%2)*2}  {counter}"
                if counter % 2:
                    prPurple("  " + out)
                else:
                    prLightPurple("  " + out)
            else:
                out = f"{line}{' '*(15-len(key) + indent*(d-2)-indent*level)} <={' '*((counter+1)%2)*2}  {counter}"
                if counter % 2:
                    prPurple("  " + out)
                else:
                    prLightPurple("  " + out)
            project.settingsList[-1].append(path)
        else:
            prRed("  " + line)
        guide.toclear += 1
    prRed(f"\n{'='*60}\n")
    guide.toclear += 6


def location(dictionary, setting, indent=6):
    prRed(f"\n{'='*60}\n")
    spots = copy.deepcopy(setting)
    d = depth(dictionary)
    text = json.dumps(dictionary, indent=indent)
    for line in text.split("\n"):
        switch = True
        level = int(((len(line) - len(line.lstrip(" "))) - indent) / indent)
        key = (
            line.replace('"', "")
            .replace("{", "")
            .replace("}", "")
            .replace(":", "")
            .replace(",", "")
            .strip()
        )
        if key:
            for path in spots:
                if not path:
                    continue
                if path[0] == key:
                    path.pop(0)
                    if not path:
                        if "," in line:
                            prYellow(
                                f"  {line}{' '*(14-len(key) + indent*(d-2)-indent*level)} <="
                            )
                        else:
                            prYellow(
                                f"  {line}{' '*(15-len(key) + indent*(d-2)-indent*level)} <="
                            )
                        switch = False
        if switch:
            prRed("  " + line)
        guide.toclear += 1
    prRed(f"\n{'='*60}\n")
    guide.toclear += 4


def pickle_unfinished(current_func):
    project.current_func = current_func
    file = os.path.join(
        current_path,
        "unfinished_config.pkl",
    )
    with open(file, "wb") as f:
        pickle.dump(project, f)


def show_settings():
    if guide.testing == False:
        return
    provars = NestedDefaultDict()
    provars["project.mode"] = project.mode
    provars["project.name"] = project.name
    provars["project.current_func"] = project.current_func
    provars["project.current_wf"] = project.current_wf
    provars["project.finished_settings"] = project.finished_settings
    provars["project.finished_maplists"] = project.finished_maplists
    provars["project.finished_set_keys"] = project.finished_set_keys
    provars["project.finished_set_maplists"] = project.finished_set_maplists
    provars["project.cores"] = project.cores
    provars["project.baseDict"] = project.baseDict
    provars["project.conditionsDict"] = project.conditionsDict
    provars["project.samplesDict"] = project.samplesDict
    provars["project.settingsDict"] = project.settingsDict
    provars["project.settingsList"] = project.settingsList
    provars["project.workflowsDict"] = project.workflowsDict
    provars["project.commentsDict"] = project.commentsDict
    print(f"============COPYME:\n")
    for name, var in provars.items():
        if isinstance(var, str):
            var = f"'{var}'"
        if isinstance(var, dict):
            var = json.dumps(var, indent=4)
        try:
            print(f"{name} = {var}")
        except:
            print(f"{name} = NestedDefaultDict()")
    print(f"\n============")


def rlinput(prompt, prefill=""):
    readline.set_startup_hook(lambda: readline.insert_text(prefill))
    try:
        return input(prompt)
    finally:
        readline.set_startup_hook()


def complete(text, state):
    return (glob.glob(text + "*") + [None])[state]


def print_intro():
    print("version: " + __version__)
    print("https://MONSDA.readthedocs.io/en/latest/index.html\n")
    print(
        bold_color.BOLD
        + "M O N S D A   C O N F I G U R A T O R"
        + bold_color.END
        + "\n"
    )


readline.set_completer_delims(" \t\n;")
readline.parse_and_bind("tab: menu-complete")
readline.set_completer(complete)


################################################################################################################
####                                            CONVERSATION                                                ####
################################################################################################################


def prepare_project(template, config):
    # add template dict to project object
    project.baseDict = template
    # add comments to commentsdict and remove them from baseDict
    comment_pathes = getPathesFromDict(project.baseDict, "comment")
    for path in comment_pathes:
        if path[-1] != "comment":
            continue
        opt = get_by_path(project.baseDict, path)
        set_by_path(project.commentsDict, path, opt)
        del_by_path(project.baseDict, path)
    show_settings()
    return intro(config)


def intro(config):
    print_intro()
    if config:
        project.mode = "modify"
        return modify(config)
    opts = {
        "1": "create project",
        "2": "create config",
        "3": "modify config",
    }
    if os.path.exists(os.path.join(current_path, "unfinished_config.pkl")):
        opts["4"] = "continue unfinished"
    guide.display(
        question="choose an option",
        options=opts,
        proof=opts.keys(),
    )
    guide.clear(3)
    if guide.answer == "1":
        project.mode = "project"
        return new_project()
    if guide.answer == "2":
        project.mode = "config"
        return new_config()
    if guide.answer == "3":
        project.mode = "modify"
        return modify()
    if guide.answer == "4":
        project.unfinished_config = os.path.join(current_path, "unfinished_config.pkl")
        return continue_unfinished()


def continue_unfinished():
    global project
    file = project.unfinished_config
    with open(project.unfinished_config, "rb") as f:
        project = pickle.load(f)
    project.unfinished_config = file
    prGreen("\nCONTINUE WITH SETTINGS FROM")
    prCyan("\n   " + project.unfinished_config)
    prRed("\n   Condition-Tree:\n")
    print_dict(project.conditionsDict, gap="      ")
    prRed("\n   Workflows:\n")
    print_dict(project.workflowsDict, gap="      ")
    prRed("\n   Settings:\n")
    print_dict(project.settingsDict, gap="      ")
    globals()[project.current_func]()


def new_config():
    # pickle_unfinished("new_config")
    prGreen("\nCREATE NEW CONFIG")
    print("\n   Name:")
    guide.display(
        question="Enter the Name of the new config file",
        proof=None,
    )
    project.name = guide.answer
    project.path = current_path
    guide.clear(4)
    prCyan(f"   Name: config_{project.name}.json")
    show_settings()
    return add_workflows()


def new_project():
    # pickle_unfinished("new_project")
    prGreen("\nCREATE NEW PROJECT")
    print("\n   Directory:")
    existing = False
    er = 0
    while True:
        if er == 0:
            ques = "Enter the absolute path where your project-folder should be created"
        if er == 1:
            ques = "couldn't find this directory"
        guide.display(question=ques, spec=current_path)
        project.name = os.path.basename(guide.answer)
        project.path = guide.answer
        if os.path.isdir(project.path) and not existing:
            guide.clear(3)
            print("")
            guide.display(
                question="WARNING: Directory already exist.",
                options={
                    "1": "enter new path",
                    "2": "add another config file to existing folder",
                },
                proof=["1", "2"],
            )
            if guide.answer == "1":
                guide.clear(6)
                continue
            if guide.answer == "2":
                guide.clear(3)
                existing = True
        if os.path.isdir(project.path) and existing:
            guide.clear(4)
            prCyan(f"   Directory: {project.path}")
            print("\n   Subname: ")
            guide.display(question=f"enter subname to differ from '{project.name}'")
            project.subname = guide.answer
            guide.clear(4)
            prCyan(f"   Subname: {project.subname}")
            break
        if os.path.isdir(os.path.dirname(project.path)):
            guide.clear(4)
            prCyan(f"   Directory: {project.path}")
            break
        else:
            guide.clear(3)
            er = 1
    show_settings()
    return add_workflows()


def add_workflows(existing_workflows=None):
    pickle_unfinished("add_workflows")
    prGreen("\nADD WORKFLOWS\n")
    possible_workflows = list(project.baseDict.keys())
    for e in none_workflow_keys:
        possible_workflows.remove(e)
    if existing_workflows:
        for e in existing_workflows:
            possible_workflows.remove(e)
    posWorkDict = NestedDefaultDict()
    counter = 1
    for w in possible_workflows:
        posWorkDict[counter] = w
        counter += 1
    guide.display(
        question="choose WORKFLOWS comma separated",
        options=posWorkDict,
        proof=list(str(i) for i in posWorkDict.keys()),
    )
    addedW = []
    for number in guide.answer.split(","):
        wf = posWorkDict[int(number)]
        addedW.append(wf)
        project.workflowsDict[wf]
    guide.clear(3)
    prCyan(f"\n   Added Workflows: {', '.join(addedW)}")
    show_settings()
    if project.mode == "modify":
        return select_conditioning()
    if project.mode == "project" or "config":
        return create_condition_tree()


def create_condition_tree():
    pickle_unfinished("create_condition_tree")
    prGreen("\nCREATE CONDITION-TREE:")
    while True:
        rec_tree_builder(project.conditionsDict, [project.name])
        print("")
        print_dict(project.conditionsDict, gap="   ")
        guide.display(
            question="press enter to continue or type 'no' to create it again",
            proof=["", "no"],
        )
        if guide.answer == "no":
            guide.toclear += 2
            guide.clear()
            project.conditionsDict = NestedDefaultDict()
        else:
            guide.toclear += 2
            guide.clear()
            # prCyan("   Condition-Tree:\n")
            print_dict(project.conditionsDict, col="Cyan", gap="         ")
            project.settingsDict = decouple(project.conditionsDict)
            break
    show_settings()
    return add_sample_dirs()


def add_sample_dirs(only_conditions=None):
    pickle_unfinished("add_sample_dirs")
    # project.current_func_arg = only_conditions
    if "FETCH" in project.workflowsDict.keys():
        return assign_SRA(only_conditions)
    print("\n  FASTQ files:")
    path_to_samples_dict = NestedDefaultDict()
    er = 0
    while True:
        if er == 0:
            ques = "Enter an absolute path where samples are stored"
            sp = current_path
        if er == 3:
            ques = "Add another path or press enter to continue"
            sp = ""
        if er == 1:
            ques = f"Sorry, couldn't find '{dir}'. Enter an absolute path where samples are stored"
            sp = dir
        if er == 2:
            ques = f"Samples must be specified to continue. Enter an absolute path where samples are stored"
            sp = current_path
        guide.display(
            question=ques,
            spec=sp,
        )
        dir = guide.answer
        if guide.answer == "" and len(project.samplesDict) == 0:
            er = 2
            guide.clear(3)
            continue
        if guide.answer == "" and len(project.samplesDict) != 0:
            guide.clear(3)
            break
        if not os.path.isdir(dir):
            er = 1
            guide.clear(3)
            continue
        if "dirpath" not in locals():
            guide.clear(4)
            prCyan("  FASTQ files:\n")
        else:
            guide.clear(3)
        print(f"         {dir}\n")
        opts = {"1": "single", "2": "paired", "3": "singlecell"}
        # print('\n')
        guide.display(
            question="Specify sequencing method", options=opts, proof=opts.keys()
        )
        seq = opts[guide.answer]
        print("")

        opts = {"1": "unstranded", "2": "rf", "3": "fr", "4": "ff"}
        guide.display(question="Specify strandedness", options=opts, proof=opts.keys())
        stranded = opts[guide.answer]

        if os.path.isdir(dir):
            counter = 0
            for dirpath, dirnames, filenames in os.walk(dir):
                for filename in [f for f in filenames if f.endswith(".fastq.gz")]:
                    counter += 1
                    name = filename.replace(".fastq.gz", "")
                    if seq == "paired":
                        if "_R1" in filename:
                            name = name.replace("_R1", "")
                        elif "_R2" in filename:
                            name = name.replace("_R2", "")
                        else:
                            prPurple(
                                f"{filename} does not follow naming convention. Please make sure all files have _R1 or respectively _R2 in their name."
                            )
                    project.samplesDict[name] = {
                        "file": filename,
                        "dir": dirpath,
                        "seq": seq,
                        "strand": stranded,
                        "cond": "",
                    }

            path_to_samples_dict[dir] = f"{counter} {seq} files found"
            guide.clear(14)
            prCyan(f"         {dir}  >  {counter} {seq} files found")
            # print_dict(project.samplesDict)
            er = 3
            continue
    counter = 1
    show_settings()
    return assign_samples(only_conditions)


def assign_SRA(only_conditions=None):
    pickle_unfinished("assign_SRA")
    prCyan("\n  Sample Assignment:  SRA Accession Numbers\n")
    print(project.mode)
    if project.mode == "modify":
        safety_dict = decouple(project.settingsDict)
    if project.mode == "project" or project.mode == "config":
        safety_dict = decouple(project.conditionsDict)
        project.settingsDict = decouple(project.conditionsDict)
    if only_conditions:
        conditions = [":".join(x) for x in only_conditions]
    else:
        conditions = [
            pattern for pattern in get_conditions_from_dict(project.conditionsDict)
        ]
    guide.toclear = 0
    while True:
        for condition in conditions:
            if (
                "SAMPLES"
                in get_by_path(project.settingsDict, condition.split(":")).keys()
            ):
                continue

            ques = f"Enter SRA Accession Numbers according to the displayed condition comma separated"
            cond_as_list = [x for x in condition.split(":")]
            location(project.settingsDict, [cond_as_list])
            guide.display(
                question=ques,
            )
            AccNumbers = guide.answer.split(",")
            set_by_path(
                project.settingsDict,
                cond_as_list,
                {"SAMPLES": AccNumbers},
            )
            guide.clear(guide.toclear + 4)
        print_dict(project.settingsDict, gap="   ")
        guide.display(
            question="press enter to continue or type 'no' to assign samples again"
        )
        if guide.answer == "no":
            guide.toclear += 2
            project.settingsDict = decouple(safety_dict)
        else:
            guide.toclear += 2
            guide.clear()
            print_dict(project.settingsDict, col="Cyan", gap="         ")
            break
    show_settings()
    return set_settings()


def assign_samples(only_conditions=None):
    pickle_unfinished("assign_samples")
    prCyan("\n  Sample Assignment:\n")
    if only_conditions != None:
        conditions = [":".join(x) for x in only_conditions]
    else:
        conditions = [
            pattern for pattern in get_conditions_from_dict(project.conditionsDict)
        ]
    if project.mode == "modify":
        safety_dict = decouple(project.settingsDict)
    if project.mode == "project" or project.mode == "config":
        safety_dict = decouple(project.conditionsDict)
        project.settingsDict = decouple(project.conditionsDict)
    er = 0
    guide.toclear = 0
    while True:
        opts = NestedDefaultDict()
        number = 1
        space = 0
        for k in project.samplesDict.keys():
            if len(k) >= space:
                space = len(k)
        if er == 0:
            for k in project.samplesDict.keys():
                opts[
                    number
                ] = f"{k}{' '*(space+2-len(k))} in  {project.samplesDict[k]['dir']}"
                number += 1
        for condition in conditions:
            if (
                "SAMPLES"
                in get_by_path(project.settingsDict, condition.split(":")).keys()
            ):
                continue
            if er == 0:
                ques = f"enter all sample-numbers according to the displayed condition comma separated or * to select all or define a range - separated"
            if er == 1:
                guide.clear(guide.toclear + 4)
                ques = f"ERROR: Its not possible to merge single-end and paired-end samples, try again"

            cond_as_list = [x for x in condition.split(":")]
            location(project.settingsDict, [cond_as_list])

            guide.display(
                question=ques, options=opts, proof=[str(i) for i in opts.keys()] + ["*"]
            )

            check_seq = set()
            if guide.answer == "*":
                for num in [str(i) for i in opts.keys()]:
                    check_seq.add(
                        project.samplesDict[opts[int(num)].split(" ")[0]]["seq"]
                    )
            else:
                for num in guide.answer.split(","):
                    if "-" in num:
                        s, e = map(int, num.split("-"))
                        for i in range(s, e + 1):
                            check_seq.add(
                                project.samplesDict[opts[int(i)].split(" ")[0]]["seq"]
                            )
                    else:
                        check_seq.add(
                            project.samplesDict[opts[int(num)].split(" ")[0]]["seq"]
                        )
            if len(check_seq) > 1:
                er = 1
                break
            else:
                er = 0
            samplesInList = []
            if guide.answer == "*":
                for num in [str(i) for i in opts.keys()]:
                    project.samplesDict[opts[int(num)].split(" ")[0]][
                        "cond"
                    ] = condition
                    samplesInList.append(opts[int(num)].split(" ")[0])
            else:
                for num in guide.answer.split(","):
                    if "-" in num:
                        s, e = map(int, num.split("-"))
                        for i in range(s, e + 1):
                            project.samplesDict[opts[int(i)].split(" ")[0]][
                                "cond"
                            ] = condition
                            samplesInList.append(opts[int(i)].split(" ")[0])
                    else:
                        project.samplesDict[opts[int(num)].split(" ")[0]][
                            "cond"
                        ] = condition
                        samplesInList.append(opts[int(num)].split(" ")[0])
            set_by_path(
                project.settingsDict,
                cond_as_list,
                {"SAMPLES": samplesInList},
            )
            guide.clear(guide.toclear + 4)
        if er == 1:
            continue
        print_dict(project.settingsDict, gap="   ")
        guide.display(
            question="press enter to continue or type 'no' to assign samples again"
        )
        if guide.answer == "no":
            guide.toclear += 2
            project.settingsDict = decouple(safety_dict)
        else:
            guide.toclear += 2
            guide.clear()
            print_dict(project.settingsDict, col="Cyan", gap="         ")
            break
    show_settings()
    return set_settings()


def set_settings():
    pickle_unfinished("set_settings")
    prGreen("\nGENERAL SETTINGS")
    print("(blank entry possible)")
    guide.toclear = 0
    safety_dict = decouple(project.settingsDict)
    conditions_list = [
        pattern.split(":")
        for pattern in get_conditions_from_dict(project.conditionsDict)
    ]
    while True:
        previous_maplist = []
        for maplist in conditions_list:
            if maplist in project.finished_set_maplists:
                continue
            prRed(f"\n   Condition:   {' : '.join(maplist)} \n")
            seq = ""
            for k in project.samplesDict.keys():
                try:
                    if project.samplesDict[k]["cond"] == ":".join(maplist):
                        seq = project.samplesDict[k]["seq"]
                except:
                    continue
            setInDict(project.settingsDict, maplist + ["SEQUENCING"], seq)
            settings_to_make = ["REFERENCE", "DECOY", "GTF", "GFF"]
            if list(set(project.workflowsDict.keys()) & set(IP_workflows)):
                settings_to_make = settings_to_make + ["IP"]
            if list(set(project.workflowsDict.keys()) & set(comparable_workflows)):
                settings_to_make = settings_to_make + ["GROUPS", "TYPES", "BATCHES"]
            if list(set(project.workflowsDict.keys()) & set(index_prefix_workflows)):
                settings_to_make = settings_to_make + ["INDEX", "PREFIX"]
            last_answer = current_path
            for key in settings_to_make:
                if key in project.finished_set_keys:
                    continue
                print(f"   {key}:")
                if key in ["REFERENCE", "DECOY", "GTF", "GFF", "IP"]:
                    if previous_maplist:
                        if key in ["GTF", "GFF"]:
                            s = get_by_path(
                                project.settingsDict,
                                previous_maplist + ["ANNOTATION", key],
                            )
                        else:
                            s = get_by_path(
                                project.settingsDict, previous_maplist + [key]
                            )
                    else:
                        s = last_answer
                    if key in ["GTF", "GFF"]:
                        p = f"end_exist_.{key.lower()}.gz"
                    elif key == "IP":
                        p = None
                    else:
                        p = "end_exist_.fa.gz"
                else:
                    p = None
                    s = None
                guide.display(
                    question=f"{project.commentsDict['SETTINGS']['comment'][key]}",
                    spec=s,
                    proof=p,
                )
                guide.clear(4)
                last_answer = guide.answer
                prCyan(f"   {key}: {guide.answer}")
                if key in ["GTF", "GFF"]:
                    setInDict(
                        project.settingsDict,
                        maplist + ["ANNOTATION", key],
                        guide.answer,
                    )
                elif key in ["GROUPS", "TYPES", "BATCHES"] and guide.answer:
                    setInDict(
                        project.settingsDict, maplist + [key], guide.answer.split(",")
                    )
                else:
                    setInDict(project.settingsDict, maplist + [key], guide.answer)
                project.finished_set_keys.append(key)
            previous_maplist = maplist
            project.finished_set_keys = []
            project.finished_set_maplists.append(maplist)
        guide.toclear = 0
        print("\n\n   SETTINGS Key:\n")
        print_dict(project.settingsDict, gap="   ")
        guide.display(
            question="\npress enter to continue or type 'no' to make settings again",
            proof=["", "no"],
        )
        if guide.answer == "":
            guide.toclear += 7
            guide.clear()
            prCyan("\n\n   SETTINGS Key:\n")
            print_dict(project.settingsDict, col="Cyan", gap="         ")
            show_settings()
            if project.mode == "project" or project.mode == "config":
                return select_conditioning()
            if project.mode == "modify":
                return fillup_workflows()
        if guide.answer == "no":
            project.settingsDict = decouple(safety_dict)
            continue


def modify(config=None):
    # pickle_unfinished("modify")
    prGreen("\nMODIFY PROJECT")
    er = 0
    while True:
        if er == 0:
            ques = "Enter the absolute path of the config file to be modified"
        if er == 1:
            ques = "couldn't find the file"
            config = None
        if er == 2:
            ques = (
                "can't read the file, check if it's the right file or if it's corrupted"
            )
        if not config:
            guide.display(question=ques, spec=current_path)
            config = guide.answer
        if not os.path.isfile(config):
            guide.clear(3)
            er = 1
            continue
        try:
            modify_config = load_configfile(config)
            project.path = os.path.dirname(config)
            project.name = list(modify_config["SETTINGS"].keys())[0]
            project.cores = modify_config["MAXTHREADS"]
            project.settingsDict = modify_config["SETTINGS"]
            project.conditionsDict = NestedDefaultDict()

            if not config:
                guide.clear(2)
            else:
                print("")
            prCyan(f"   {config}")
        except:
            er = 2

        condition_pathes = getPathesFromDict(modify_config, "SAMPLES")
        for path in condition_pathes:
            if path[-1] == "SAMPLES":
                path = path[1:-1]
                setInDict(project.conditionsDict, path, {})
        project.finished_set_maplists = [
            pattern.split(":")
            for pattern in get_conditions_from_dict(project.conditionsDict)
        ]
        active_workflows = modify_config["WORKFLOWS"].split(",")
        inactive_workflows = list(modify_config.keys())
        for e in none_workflow_keys:
            inactive_workflows.remove(e)
        for wf in inactive_workflows:
            project.workflowsDict[wf] = decouple(modify_config[wf])
        for e in modify_config["WORKFLOWS"].split(","):
            inactive_workflows.remove(e)
        prRed("\n   Following configuration was found :\n")
        prRed("   Condition-Tree:\n")
        print_dict(project.conditionsDict, gap="      ")
        prRed("\n   active WORKFLOWS:\n")
        if active_workflows:
            print("      " + ", ".join(active_workflows))
        else:
            print("      -")
        prRed("\n   inactive WORKFLOWS:\n")
        if inactive_workflows:
            print("      " + ", ".join(inactive_workflows))
        else:
            print("      -")

        print("\nWhat do you want to do...\n")
        opts = NestedDefaultDict()
        opts["1"] = "add workflows"
        opts["2"] = "remove workflows"
        opts["3"] = "add conditions"
        opts["4"] = "remove conditions"
        opts["5"] = "choose another file"
        guide.display(question="choose an option", options=opts, proof=opts.keys())
        guide.clear(3)
        if guide.answer == "1":
            return add_workflows(inactive_workflows + active_workflows)
        if guide.answer == "2":
            return remove_workflows(active_workflows, inactive_workflows)
        if guide.answer == "3":
            return add_conditions()
        if guide.answer == "4":
            return remove_conditions()
        if guide.answer == "5":
            project.settingsDict = NestedDefaultDict()
            project.conditionsDict = NestedDefaultDict()
            project.workflowsDict = NestedDefaultDict()
            continue


def add_conditions():
    pickle_unfinished("add_conditions")
    prGreen("\n\nADD CONDIIONS")
    old_conditions = [
        pattern.split(":")
        for pattern in get_conditions_from_dict(project.conditionsDict)
    ]
    while True:
        new_conditions = decouple(project.conditionsDict)
        rec_tree_builder(new_conditions, [project.name], add_mode=True)
        print("\n   New Condition-Tree:\n")
        print_dict(new_conditions, gap="   ")
        guide.display(
            question="press enter to apply changes or type 'no' to make changes again",
            proof=["", "no"],
        )
        if guide.answer == "no":
            guide.toclear += 3
            guide.clear()
            continue
        else:
            project.conditionsDict = new_conditions
            guide.toclear += 4
            guide.clear()
            prCyan("   New Condition-Tree:\n")
            print_dict(project.conditionsDict, col="Cyan", gap="         ")
            # project.settingsDict = decouple(project.conditionsDict)
            break

    all_conditions = [
        pattern.split(":")
        for pattern in get_conditions_from_dict(project.conditionsDict)
    ]
    print(all_conditions)
    new_conditions = [x for x in all_conditions if x not in old_conditions]
    print(new_conditions)
    for condition in new_conditions:
        setInDict(project.settingsDict, condition, NestedDefaultDict())
        for wf in project.workflowsDict.keys():
            setInDict(project.workflowsDict, [wf] + condition, NestedDefaultDict())
    show_settings()
    return add_sample_dirs(new_conditions)


def remove_workflows(actives, inactives):
    pickle_unfinished("remove_workflows")
    prGreen("\n\nREMOVE WORKFLOWS")
    workflows = project.workflowsDict.keys()
    opts = NestedDefaultDict()
    number = 1
    for wf in workflows:
        opts[str(number)] = wf
        number += 1
    while True:
        print("\nfollowing workflows exist:\n")
        guide.display(
            question="enter numbers of workflows to be removed comma separated",
            options=opts,
            proof=opts.keys(),
        )
        new_workflows = decouple(project.workflowsDict)
        new_actives = copy.deepcopy(actives)
        new_inactives = copy.deepcopy(inactives)
        for key, value in opts.items():
            if key in guide.answer.split(","):
                new_workflows.pop(value)
                if value in new_actives:
                    new_actives.remove(value)
                if value in new_inactives:
                    new_inactives.remove(value)
        prRed("\n   active WORKFLOWS:\n")
        if new_actives:
            print("      " + ", ".join(new_actives))
        else:
            print("      -")
        prRed("\n   inactive WORKFLOWS:\n")
        if new_inactives:
            print("      " + ", ".join(new_inactives))
        else:
            print("      -")
        guide.display(
            question="press enter to apply changes or type 'no' to make changes again",
            proof=["", "no"],
        )
        if guide.answer == "":
            project.workflowsDict = new_workflows
            return finalize()
        if guide.answer == "no":
            continue


def remove_conditions():
    pickle_unfinished("remove_conditions")
    prGreen("\n\nREMOVE CONDITIONS")
    conditions = get_conditions_from_dict(project.conditionsDict)
    opts = NestedDefaultDict()
    number = 1
    for c in conditions:
        opts[str(number)] = c
        number += 1
    while True:
        new_settings = decouple(project.settingsDict)
        new_workflows = decouple(project.workflowsDict)
        new_conditions = decouple(project.conditionsDict)
        print("\nfollowing conditions exist:\n")
        guide.display(
            question="enter numbers of conditions to be removed comma separated",
            options=opts,
            proof=opts.keys(),
        )
        numbers = guide.answer.split(",")
        for number in numbers:
            path_to_remove = opts[number].split(":")
            del_by_path(new_settings, path_to_remove)
            del_by_path(new_conditions, path_to_remove)
            for key in new_workflows.keys():
                del_by_path(new_workflows, [key] + path_to_remove)
        prRed("\nNew condition-Tree:")
        print_dict(new_conditions)
        guide.display(
            question="press enter to apply changes or type 'no' to make changes again",
            proof=["", "no"],
        )
        if guide.answer == "":
            project.settingsDict = new_settings
            project.workflowsDict = new_workflows
            project.conditionsDict = new_conditions
            return finalize()
        if guide.answer == "no":
            continue


def select_conditioning():
    pickle_unfinished("select_conditioning")
    conditions = [
        pattern.split(":")
        for pattern in get_conditions_from_dict(project.conditionsDict)
    ]
    if len(conditions) == 1:
        project.settingsList = [conditions]
        return set_workflows()
    prGreen("\n\nSELECT CONDITIONS")
    guide.toclear = 0
    while True:
        d = depth(project.conditionsDict)
        for i in range(d - 1):
            select_id_to_set(project.conditionsDict, i)
            print(
                "In the following steps you will make different settings for each condition.\nTo avoid repetitions, specify which conditions should get the same settings\nYou will set all conditions with the same number at once afterwards"
            )
            guide.display(
                question="To loop through the possible selections press enter\n Type 'ok' to accept your settings",
                proof=["ok", ""],
            )
            guide.toclear += 6
            if guide.answer == "ok":
                guide.clear(8)
                show_settings()
                return set_workflows()
            else:
                guide.clear()


def fillup_workflows():
    pickle_unfinished("fillup_workflows")
    conditions = [
        pattern.split(":")
        for pattern in get_conditions_from_dict(project.conditionsDict)
    ]
    fillupDict = NestedDefaultDict()
    for wf in project.workflowsDict.keys():
        fillupDict[wf]["on"] = []
        fillupDict[wf]["off"] = []
        for condition in conditions:
            if not get_by_path(project.workflowsDict, [wf] + condition):
                fillupDict[wf]["off"].append(condition)
            if get_by_path(project.workflowsDict, [wf] + condition):
                fillupDict[wf]["on"].append(condition)

    for wf in project.workflowsDict.keys():
        for off_con in fillupDict[wf]["off"]:
            location(project.workflowsDict[wf], [off_con])
            prGreen(f"FILLUP WORKFLOW-SETTINGS FOR {wf}\n")
            opts = NestedDefaultDict()
            number = 1
            for on_con in fillupDict[wf]["on"]:
                opts[str(number)] = f"copy from {':'.join(on_con)}"
                number += 1
            opts[str(number)] = "make new settings"
            guide.display(question="enter option", options=opts, proof=opts.keys())
            if opts[guide.answer] == "make new settings":
                project.settingsList = [[off_con]]
                set_workflows(wf)
            else:
                copy_con = opts[guide.answer].split(" ")[-1].split(":")
                toadd = get_by_path(project.workflowsDict, [wf] + copy_con)
                setInDict(project.workflowsDict, [wf] + off_con, toadd)
    return finalize()


def set_workflows(wf=None):
    pickle_unfinished("set_workflows")
    if wf:
        workflows = [wf]
    else:
        workflows = project.workflowsDict.keys()
    for workflow in workflows:
        if not project.workflowsDict[workflow] or wf or workflow == project.current_wf:
            project.current_wf = workflow
            prGreen(f"\nMAKE SETTINGS FOR {workflow}\n")
            opt_dict = NestedDefaultDict()
            tools_to_use = NestedDefaultDict()
            features_to_use = NestedDefaultDict()
            #
            if (
                "TOOLS" in project.baseDict[workflow].keys()
                and project.workflowsDict[workflow]["TOOLS"]
            ):
                tools_to_use = project.workflowsDict[workflow]["TOOLS"]
            if (
                "TOOLS" in project.baseDict[workflow].keys()
                and not project.workflowsDict[workflow]["TOOLS"]
            ):
                number = 1
                for k in project.baseDict[workflow]["TOOLS"].keys():
                    opt_dict[number] = k
                    number += 1
                if number > 2:
                    print(f"   Tools:\n")
                    guide.display(
                        question="Select from these available Tools comma separated:",
                        options=opt_dict,
                        proof=[str(i) for i in opt_dict.keys()],
                    )
                    guide.clear(len(opt_dict) + 5)
                    for number in guide.answer.split(","):
                        tools_to_use[opt_dict[int(number)]] = project.baseDict[
                            workflow
                        ]["TOOLS"][opt_dict[int(number)]]
                    prCyan(f"   Tools: {', '.join(tools_to_use.keys())}")
                else:
                    tools_to_use[opt_dict[int(1)]] = project.baseDict[workflow][
                        "TOOLS"
                    ][opt_dict[int(1)]]
                    prCyan(f"   Tool: {', '.join(tools_to_use.keys())}")

            if (
                "FEATURES" in project.baseDict[workflow].keys()
                and project.workflowsDict[workflow]["FEATURES"]
            ):
                features_to_use = project.workflowsDict[workflow]["FEATURES"]
            if (
                "FEATURES" in project.baseDict[workflow].keys()
                and not project.workflowsDict[workflow]["FEATURES"]
            ):
                number = 1
                for k in project.baseDict[workflow]["FEATURES"].keys():
                    opt_dict[number] = k
                    number += 1
                if number > 2:
                    print(f"   FEATURES:\n")
                    guide.display(
                        question='Select from these available FEATURES comma separated. Defines FEATURES that are used to count as key and their identifiers in the corresponding annotation as value, e.g "gene":"gene_id". IFF other features than those preselected are needed or they do not map to the standard identifiers (gene_id for exon and gene, please change them manually after config was created.)',
                        options=opt_dict,
                        proof=[str(i) for i in opt_dict.keys()],
                    )
                    guide.clear(len(opt_dict) + 5)
                    for number in guide.answer.split(","):
                        features_to_use[opt_dict[int(number)]] = project.baseDict[
                            workflow
                        ]["FEATURES"][opt_dict[int(number)]]
                    prCyan(f"   FEATURES: {', '.join(features_to_use.keys())}")
                else:
                    features_to_use[opt_dict[int(1)]] = project.baseDict[workflow][
                        "FEATURES"
                    ][opt_dict[int(1)]]
                    prCyan(f"   FEATURE: {', '.join(features_to_use.keys())}")

            if (
                workflow == "PEAKS"
                and "macs" in ", ".join(tools_to_use.keys())
                or workflow in comparable_workflows
            ):
                prRed(f"\n   Settings for differential Analyses:\n")

            if "CUTOFFS" in project.baseDict[workflow].keys() and not wf:
                project.workflowsDict[workflow].update({"CUTOFFS": {}})
                project.workflowsDict[workflow]["CUTOFFS"].update(
                    project.baseDict[workflow]["CUTOFFS"]
                )

                for key, value in project.baseDict[workflow]["CUTOFFS"].items():
                    print(f"      {key}:")
                    guide.display(
                        question=f"set cutoff", proof="only_numbers", spec=value
                    )
                    guide.clear(4)
                    prCyan(f"      {key}: {guide.answer}")
                    set_by_path(
                        project.workflowsDict,
                        [workflow, "CUTOFFS", key],
                        str(guide.answer),
                    )

            if (
                workflow == "PEAKS"
                and "macs" in ", ".join(tools_to_use.keys())
                or workflow in comparable_workflows
                and not project.workflowsDict[workflow]["COMPARABLE"]
            ):
                project.workflowsDict[workflow]["COMPARABLE"]
                project.workflowsDict[workflow]["EXCLUDE"] = []
                prCyan(f"      Comparables:\n")
                groups = set()
                conditions = get_conditions_from_dict(project.conditionsDict)
                for condition in conditions:
                    groups.update(
                        get_by_path(
                            project.settingsDict, condition.split(":") + ["GROUPS"]
                        )
                    )
                if len(groups) < 2:
                    guide.clear(2)
                    prCyan(f"      Comparables: No Groups found\n")
                    project.workflowsDict[workflow]["COMPARABLE"] = ""
                else:
                    while True:
                        guide.display(
                            question="Press enter to add another differential contrast or type 'ok' to accept",
                            proof=["", "ok"],
                        )
                        if guide.answer == "ok":
                            guide.clear(2)
                            break

                        number = 1
                        guide.clear(2)
                        opt_dict = NestedDefaultDict()
                        for g in sorted(groups):
                            opt_dict[str(number)] = g
                            number += 1
                        contrast = [[], []]
                        guide.display(
                            question="select base-condition",
                            options=opt_dict,
                            proof=opt_dict.keys(),
                        )
                        guide.toclear += 2
                        contrast[0].append(opt_dict[guide.answer])
                        del opt_dict[guide.answer]
                        print("")
                        guide.display(
                            question="select contrast-condition",
                            options=opt_dict,
                            proof=opt_dict.keys(),
                        )
                        guide.toclear += 2
                        contrast[1].append(opt_dict[guide.answer])

                        guide.toclear += 2
                        guide.clear()
                        comp_name = "-VS-".join(x[0] for x in contrast)
                        prCyan(f"         {comp_name}")
                        project.workflowsDict[workflow]["COMPARABLE"][
                            comp_name
                        ] = contrast
                guide.clear(1)

            for setting in project.settingsList:
                if setting in project.finished_settings:
                    continue
                prRed(
                    f"\n   Selected Conditions:   {' / '.join([':'.join(c) for c in setting])}"
                )
                for tool, bin in tools_to_use.items():
                    project.workflowsDict[workflow]["TOOLS"][tool] = bin
                    for maplist in setting:
                        if maplist in project.finished_maplists:
                            continue
                        setInDict(
                            project.workflowsDict,
                            [workflow] + maplist + [tool, "OPTIONS"],
                            {},
                        )

                    prPurple(f"\n      Set OPTIONS for tool:  {tool}\n")
                    for option in project.baseDict[workflow][tool]["OPTIONS"]:
                        print(f"      Option: '{option}'")
                        call = project.baseDict[workflow][tool]["OPTIONS"][option]
                        guide.toclear = 0
                        guide.display(
                            question=f"{project.commentsDict[workflow][tool]['comment'][option]}",
                            spec=call,
                            whitespace=True,
                        )
                        optsDict = guide.answer
                        for maplist in setting:
                            setInDict(
                                project.workflowsDict,
                                [workflow] + maplist + [tool, "OPTIONS", option],
                                optsDict,
                            )
                        project.finished_maplists.append(maplist)
                        guide.clear(4)
                        prCyan(f"      {option}:  {guide.answer}")
                for feat, fid in features_to_use.items():
                    project.workflowsDict[workflow]["FEATURES"][feat] = fid
                project.finished_settings.append(setting)
                project.finished_maplists = []
            project.finished_settings = []
    show_settings()
    if project.mode == "project" or project.mode == "config":
        return set_cores()
    if project.mode == "modify":
        return finalize()


def set_cores():
    pickle_unfinished("set_cores")
    prGreen("\nSET THREADS\n")
    print("   MAXTHREADS:")
    guide.display(
        question="set the maximum number of cores that can be used",
        proof="only_numbers",
    )
    guide.clear(5)
    prCyan(f"\n   MAXTHREADS: {guide.answer}")
    project.cores = str(guide.answer)
    show_settings()
    return finalize()


def finalize():
    pickle_unfinished("finalize")
    final_dict = NestedDefaultDict()

    final_dict["WORKFLOWS"] = ",".join(project.workflowsDict.keys())
    final_dict["BINS"] = ""  # project.baseDict["BINS"]
    final_dict["MAXTHREADS"] = project.cores
    final_dict["VERSION"] = __version__
    final_dict["SETTINGS"] = project.settingsDict
    final_dict.update(project.workflowsDict)

    if project.subname:
        configfile = f"config_{'_'.join([project.name,project.subname])}.json"
    else:
        configfile = f"config_{project.name}.json"
    print("\n")
    prGreen("FINAL CONFIGURATION:\n")
    print_dict(final_dict)
    print("\n")

    if project.mode == "config":
        print("Above is your final configuration of MONSDA.")
        guide.display(
            question=f"\npress enter to create the config_{project.name}.json file or type 'abort' before it gets serious",
            proof=["", "abort"],
        )
    if project.mode == "project":
        space = len(configfile)
        print(
            "Above is your final configuration of MONSDA. The Guide will create this directory as new project:\n"
        )
        prGreen(f"  {os.path.dirname(project.path)}")
        prGreen(f"  └─{os.path.basename(project.path)}")
        prGreen(
            f"     ├─FASTQ{' '*(space-5)}   >  contains symlinks of your samplefiles"
        )
        prGreen(
            f"     ├─GENOMES{' '*(space-7)}   >  contains symlinks of your reference files"
        )
        prGreen(f"     └─{configfile}   >  your brand new configuration file")

        guide.display(
            question="\npress enter to create your project or type 'abort' before it gets serious",
            proof=["", "abort"],
        )
    if project.mode == "modify":
        print("Above is your updated configuration of MONSDA\n")
        print(
            "The old config-file will be preserved (it will have a timestamp in it's name)\n"
        )
        guide.display(
            question=f"\npress enter to update {configfile} or type 'abort' before it gets serious",
            proof=["", "abort"],
        )

    if guide.answer == "abort":
        quit()
    else:
        guide.clear(3)
        return create_project(final_dict)


def create_project(final_dict):
    if project.mode == "project":
        # create Project Folder
        cwd = os.getcwd()
        if not os.path.exists(project.path):
            os.mkdir(project.path)
        fastq = os.path.join(project.path, "FASTQ", project.name)
        gen = os.path.join(project.path, "GENOMES")
        if not os.path.exists(fastq):
            os.makedirs(fastq)
        if not os.path.exists(gen):
            os.makedirs(gen)

        # LINK samples into FASTQ and insert samplenames in dict
        for k in project.samplesDict.keys():
            if project.samplesDict[k]["cond"]:
                condition = project.samplesDict[k]["cond"]
                cond_as_list = [x for x in condition.split(":")[1:]]
                os.chdir(fastq)
                for dir in cond_as_list:
                    if not os.path.exists(os.path.join(dir)):
                        os.mkdir(os.path.join(dir))
                    os.chdir(os.path.join(dir))
                path = "/".join(cond_as_list)
                cond_dir = os.path.join(fastq, path)
                dst = os.path.join(cond_dir, project.samplesDict[k]["file"])
                src = os.path.join(
                    project.samplesDict[k]["dir"], project.samplesDict[k]["file"]
                )
                if not os.path.exists(dst):
                    os.symlink(src, dst)
                    if "paired" in project.samplesDict[k]["seq"]:
                        if "_R1" in project.samplesDict[k]["file"]:
                            other = project.samplesDict[k]["file"].replace("_R1", "_R2")
                        elif "_R2" in project.samplesDict[k]["file"]:
                            other = project.samplesDict[k]["file"].replace("_R2", "_R1")
                        else:
                            prRed(
                                f"WARNING: Paired sample {project.samplesDict[k]} isn't named correctly. Check R1, R2 in filename"
                            )
                        dst = os.path.join(cond_dir, other)
                        src = os.path.join(project.samplesDict[k]["dir"], other)
                        os.symlink(src, dst)

        # link reference and annotation
        for setting in project.settingsList:
            for condition in setting:
                ref = get_by_path(project.settingsDict, condition + ["REFERENCE"])
                if os.path.isfile(ref):
                    if not os.path.exists(os.path.join(gen, os.path.basename(ref))):
                        os.symlink(
                            os.path.realpath(ref),
                            os.path.join(gen, os.path.basename(ref)),
                        )
                    f = os.path.join(gen, os.path.basename(ref))
                    rel = os.path.relpath(f, start=project.path)
                    setInDict(project.settingsDict, condition + ["REFERENCE"], rel)
                else:
                    prRed(
                        f"WARNING: reference path at {condition} is not correct, could not symlink, please do by hand"
                    )
                    setInDict(project.settingsDict, condition + ["REFERENCE"], "")

                decoy = get_by_path(project.settingsDict, condition + ["DECOY"])
                if os.path.isfile(decoy):
                    if not os.path.exists(os.path.join(gen, os.path.basename(decoy))):
                        os.symlink(
                            os.path.realpath(decoy),
                            os.path.join(gen, os.path.basename(decoy)),
                        )
                    f = os.path.join(gen, os.path.basename(decoy))
                    rel = os.path.relpath(f, start=project.path)
                    setInDict(project.settingsDict, condition + ["DECOY"], rel)
                else:
                    prRed(
                        f"WARNING: decoy path at {condition} is not correct, could not symlink, please do by hand if needed"
                    )
                    setInDict(project.settingsDict, condition + ["DECOY"], "")

                gtf = get_by_path(
                    project.settingsDict, condition + ["ANNOTATION", "GTF"]
                )
                if os.path.isfile(gtf):
                    if not os.path.exists(os.path.join(gen, os.path.basename(gtf))):
                        os.symlink(
                            os.path.realpath(gtf),
                            os.path.join(gen, os.path.basename(gtf)),
                        )
                    f = os.path.join(gen, os.path.basename(gtf))
                    rel = os.path.os.path.relpath(f, start=project.path)
                    setInDict(
                        project.settingsDict, condition + ["ANNOTATION", "GTF"], rel
                    )
                else:
                    prRed(
                        f"WARNING: GTF path at {condition} is not correct or not given, could not symlink"
                    )
                    setInDict(
                        project.settingsDict, condition + ["ANNOTATION", "GTF"], ""
                    )

                gff = get_by_path(
                    project.settingsDict, condition + ["ANNOTATION", "GFF"]
                )
                if os.path.isfile(gff):
                    if not os.path.exists(os.path.join(gen, os.path.basename(gff))):
                        os.symlink(
                            os.path.realpath(gff),
                            os.path.join(gen, os.path.basename(gff)),
                        )
                    f = os.path.join(gen, os.path.basename(gff))
                    rel = os.path.os.path.relpath(f, start=project.path)
                    setInDict(
                        project.settingsDict, condition + ["ANNOTATION", "GFF"], rel
                    )
                else:
                    prRed(
                        f"WARNING: GFF path at {condition} is not correct or not given, could not symlink"
                    )
                    setInDict(
                        project.settingsDict, condition + ["ANNOTATION", "GFF"], ""
                    )

    if project.mode == "modify":
        file = os.path.join(project.path, f"config_{project.name}.json")
        bakfile = os.path.join(
            project.path,
            f"config_{project.name}_{datetime.datetime.now().strftime('%Y%m%d_%H_%M_%S')}.json",
        )
        os.system(f"cp {file} {bakfile}")

    if project.subname:
        configfile = f"config_{'_'.join([project.name,project.subname])}.json"
    else:
        configfile = f"config_{project.name}.json"
    with open(os.path.join(project.path, configfile), "w") as jsonout:
        print(json.dumps(final_dict, indent=4), file=jsonout)
    os.remove(os.path.join(current_path, project.unfinished_config))
    print(f"\nStart MONSDA with\n")
    prGreen(f"   monsda -c {configfile} --directory ${{PWD}}\n\n")


def main():
    global project
    global guide
    project = PROJECT()
    guide = GUIDE()
    if args.test:
        guide.testing = True
    if args.config:
        if not args.config.endswith(".json"):
            print("Session flag requires .json file")
            exit()
        file = os.path.join(current_path, args.config)
        config = file
    else:
        config = None
    if args.session:
        if not args.session.endswith(".pkl"):
            print("Session flag requires .pkl file")
            exit()
        project.unfinished_config = os.path.join(current_path, args.session)
        print_intro()
        continue_unfinished()
    else:
        prepare_project(template, config)


####################
####    MAIN    ####
####################

if __name__ == "__main__":
    main()



# GenerateTrackDb.py ---
#
# Filename: GenerateTrackDb.py
# Description:
# Author: Joerg Fallmann
# Maintainer:
# Created: Mon Dec  4 09:54:46 2017 (+0100)
# Version:
# Package-Requires: ()
# Last-Updated: Mon Mar  9 10:23:08 2020 (+0100)
#           By: Joerg Fallmann
#     Update #: 161
# URL:
# Doc URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#
#

# Change Log:
#
#
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
#
#

# Code:
###############
###Imports
###############
import sys
import os
import glob
import time
import gzip
import string
import argparse
import random
import traceback as tb
import inspect
import logging

cmd_subfolder = os.path.join(
    os.path.dirname(
        os.path.realpath(os.path.abspath(inspect.getfile(inspect.currentframe())))
    ),
    "../lib",
)
if cmd_subfolder not in sys.path:
    sys.path.insert(0, cmd_subfolder)

from Logger import *

###############
###Variables
###############

verbose = 0
date = time.strftime("%d/%m/%Y")

try:
    scriptname = os.path.basename(inspect.stack()[-1].filename).replace(".py", "")
    log = logging.getLogger(scriptname)

    lvl = log.level if log.level else "INFO"
    for handler in log.handlers[:]:
        handler.close()
        log.removeHandler(handler)

    handler = logging.FileHandler("LOGS/MONSDA.log", mode="a")
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    log.setLevel(lvl)

except Exception:
    exc_type, exc_value, exc_tb = sys.exc_info()
    tbe = tb.TracebackException(
        exc_type,
        exc_value,
        exc_tb,
    )
    print("".join(tbe.format()), file=sys.stderr)

###############
###Command Line Options
###############


def parseargs():
    parser = argparse.ArgumentParser(description="GenerateTrackDb.py")
    parser.add_argument(
        "-e",
        "--header",
        type=bool,
        default=False,
        help="If header should be created or just entries added to the trackdb file",
    )
    parser.add_argument(
        "-t",
        "--track",
        type=str,
        default="trackdb.txt",
        help="Name of the trackdb file to generate or to extend",
    )
    parser.add_argument(
        "-f",
        "--filenames",
        type=str,
        help="Name of the file containing bwfilenames that should be added to the trackdb file, can be STDIN",
    )
    parser.add_argument(
        "-n",
        "--name",
        type=str,
        default="Autogentrack",
        help="Name of the track to create, can be anything",
    )
    parser.add_argument(
        "-s",
        "--shortlabel",
        type=str,
        help="Short label for track and tracks, default is first 20chars of bw files",
    )
    parser.add_argument(
        "-l",
        "--longlabel",
        type=str,
        help="Long label for track and tracks, default is name of bw files",
    )
    parser.add_argument(
        "-u",
        "--baseurl",
        type=str,
        help="Base url that will be used for all the tracks",
    )
    parser.add_argument(
        "-g",
        "--genome",
        type=str,
        help="Genome used for analysis, will be added to genomes.txt",
    )
    parser.add_argument(
        "-b",
        "--hub",
        type=str,
        default="AutogenHub",
        help="Will generate hub.txt file.",
    )
    parser.add_argument(
        "-m", "--mail", type=str, default="is@egal.com", help="Email of track owner."
    )
    parser.add_argument(
        "-i",
        "--uid",
        type=str,
        default="",
        help="Unique ID for output in case of bulk generation",
    )
    parser.add_argument(
        "-x",
        "--splitbw",
        type=bool,
        default=False,
        help="Will split into reverse and forward tracks, default is to negate reverse strand values.",
    )
    return parser.parse_args()


###############
###MAIN
###############


def main(
    header,
    track,
    shortlabel,
    longlabel,
    filenames,
    name,
    url,
    genome,
    hub,
    mail,
    splitbw,
    uid,
):
    try:
        if header and splitbw:
            if not shortlabel:
                shortlabel = os.path.split(name)[1].split(".")[0]
            if not longlabel:
                longlabel = name

            head = "\n".join(
                [
                    "# TrackDb_generated_by_GenerateTrackDb.txt",
                    "track {tr}",
                    "container multiWig",
                    "noInherit on",
                    "shortLabel {sl}",
                    "longLabel {ll}",
                    "type bigWig",
                    "configurable on",
                    "visibility full",
                    "aggregate transparentOverlay",
                    "showSubtrackColorOnUi on",
                    "autoScale on",
                    "windowingFunction maximum",
                    "priority 1000",
                    "alwaysZero on",
                    "yLineMark 0",
                    "yLineOnOff on",
                    "maxHeightPixels 200:64:32",
                ]
            ).format(tr=name, sl=shortlabel, ll=longlabel)

            with open(uid + "re_" + track, "a") as out:
                out.write(head + "\n\n")
            with open(uid + "fw_" + track, "a") as out:
                out.write(head + "\n\n")

        if filenames == "STDIN" or filenames == "-":
            bigwigs = []
            for line in sys.stdin:
                bigwigs.append(line.split("\n", 2)[0])

        elif os.path.isfile(filenames):
            bigwigs = []
            if ".gz" in filenames:
                b = gzip.open(filenames, "rb")
                for line in b:
                    bigwigs.append(line.split("\n", 2)[0])
                    b.close()
            else:
                b = open(filenames, "r")
                for line in b:
                    bigwigs.append(line.split("\n", 2)[0])
                    b.close()

        fwtracks = []
        retracks = []
        head = ""
        idx = 1

        for content in bigwigs:
            bw = content
            sample = os.path.split(bw)[1].split(".")[0]
            colorfw = "128,0,0"  # random_color()
            colorre = "0,128,0"
            shortlabel = sample
            longlabel = bw.split(".bw", 2)[0]

            if "\t" in bw:
                bw, sample = content.split("\t", 2)

            if header:
                if not splitbw and ".re" in longlabel:
                    content = "\n".join(
                        [
                            "track " + sample + "_re",
                            "bigDataUrl {1}/" + bw,
                            "shortLabel {2}",
                            "longLabel {3}",
                            "type bigWig",
                            "negateValues on #uncomment of this is wanted",
                            "parent {4}",
                            "color {5}",
                        ]
                    ).format(bw, url, shortlabel, longlabel, sample, colorre)
                elif not splitbw and not ".re" in longlabel:
                    head = "\n".join(
                        [
                            "# TrackDb_generated_by_GenerateTrackDb.txt",
                            "track {tr}",
                            "container multiWig",
                            "noInherit on",
                            "shortLabel {sl}",
                            "longLabel {ll}",
                            "type bigWig",
                            "configurable on",
                            "visibility full",
                            "aggregate transparentOverlay",
                            "showSubtrackColorOnUi on",
                            "autoScale on",
                            "windowingFunction maximum",
                            "priority 1000",
                            "alwaysZero on",
                            "yLineMark 0",
                            "yLineOnOff on",
                            "maxHeightPixels 200:64:32",
                        ]
                    ).format(tr=sample, sl=shortlabel, ll=longlabel)

                    content = "\n".join(
                        [
                            "track " + sample + "_fw",
                            "bigDataUrl {1}/" + bw,
                            "shortLabel {2}",
                            "longLabel {3}",
                            "type bigWig",
                            "#negateValues on #uncomment of this is wanted",
                            "parent {4}",
                            "color {5}",
                        ]
                    ).format(bw, url, shortlabel, longlabel, sample, colorfw)
            else:
                if not splitbw and ".re" in longlabel:
                    content = "\n".join(
                        [
                            "track " + sample + "_re" + str(idx),
                            "bigDataUrl {1}/" + bw,
                            "shortLabel {2}" + str(idx),
                            "longLabel {3}" + str(idx),
                            "type bigWig",
                            "negateValues on #uncomment of this is wanted",
                            "color {5}",
                        ]
                    ).format(bw, url, shortlabel, longlabel, name, colorre)
                else:
                    content = "\n".join(
                        [
                            "track " + sample + "_fw" + str(idx),
                            "bigDataUrl {1}/" + bw,
                            "shortLabel {2}" + str(idx),
                            "longLabel {3}" + str(idx),
                            "type bigWig",
                            "#negateValues on #uncomment of this is wanted",
                            "color {5}",
                        ]
                    ).format(bw, url, shortlabel, longlabel, name, colorfw)

            idx += 1

            if splitbw and ".re" in longlabel:
                retracks.append(content)
            elif ".re" in longlabel:
                fwtracks.append(content)
            else:
                fwtracks.append(head + "\n\n" + content)

        if len(retracks) > 0:
            for i in retracks:
                with open(uid + "re_" + track, "a") as out:
                    out.write(i + "\n\n")

            for i in fwtracks:
                with open(uid + "fw_" + track, "a") as out:
                    out.write(i + "\n\n")

        else:
            for i in fwtracks:
                with open(uid + track, "a") as out:
                    out.write(i + "\n\n")

        if genome:
            if retracks:
                gen = "\n".join(
                    ["genome " + genome, "trackDb " + uid + "fw_{0}"]
                ).format(track)
                with open(uid + "genomes_fw.txt", "w") as out:
                    out.write(gen + "\n")

                gen = "\n".join(
                    ["genome " + genome, "trackDb " + uid + "re_{0}"]
                ).format(track)
                with open(uid + "genomes_re.txt", "w") as out:
                    out.write(gen + "\n")

            else:
                gen = "\n".join(["genome " + genome, "trackDb " + uid + "{0}"]).format(
                    track
                )
                with open(uid + "genomes.txt", "w") as out:
                    out.write(gen + "\n")

        if hub:
            if retracks:
                hubtxt = "\n".join(
                    [
                        "hub fw_" + hub,
                        "shortLabel FW_{0}",
                        "longLabel FW_{1}",
                        "genomesFile " + uid + "genomes_fw.txt",
                        "email {2}",
                    ]
                ).format(shortlabel, longlabel, mail)
                with open(uid + "hub_fw.txt", "w") as out:
                    out.write(hubtxt + "\n")

                hubtxt = "\n".join(
                    [
                        "hub re_" + hub,
                        "shortLabel RE_{0}",
                        "longLabel RE_{1}",
                        "genomesFile " + uid + "genomes_re.txt",
                        "email {2}",
                    ]
                ).format(shortlabel, longlabel, mail)
                with open(uid + "hub_re.txt", "w") as out:
                    out.write(hubtxt + "\n")

            else:
                hubtxt = "\n".join(
                    [
                        "hub " + hub,
                        "shortLabel {0}",
                        "longLabel {1}",
                        "genomesFile " + uid + "genomes.txt",
                        "email {2}",
                    ]
                ).format(shortlabel, longlabel, mail)
                with open(uid + "hub.txt", "w") as out:
                    out.write(hubtxt + "\n")
    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        print("".join(tbe.format()))


###############
###Subs
###############
def random_color():
    rgbl = [255, 0, 0]
    random.shuffle(rgbl)
    return str(rgbl)


###############
###Standalone
###############

if __name__ == "__main__":
    args = parseargs()
    main(
        args.header,
        args.track,
        args.shortlabel,
        args.longlabel,
        args.filenames,
        args.name,
        args.baseurl,
        args.genome,
        args.hub,
        args.mail,
        args.splitbw,
        args.uid,
    )

##################################END################################


#
# GenerateTrackDb.py ends here



#!/usr/bin/env python3

import argparse
import os
import shlex
import shutil
import subprocess
import sys
import traceback as tb

from snakemake.common.configfile import load_configfile

# from snakemake.utils import min_version

scriptname = os.path.basename(__file__).replace("Run", "").replace(".py", "")

from MONSDA.Logger import makelogdir, setup_logger

from . import _version

__version__ = _version.get_versions()["version"]

# Logging
import datetime

makelogdir("LOGS")
if not os.path.isfile(os.path.abspath("LOGS" + os.sep + scriptname + ".log")):
    open("LOGS/" + scriptname + ".log", "w").close()
else:
    ts = str(
        datetime.datetime.fromtimestamp(
            os.path.getmtime(os.path.abspath("LOGS" + os.sep + scriptname + ".log"))
        ).strftime("%Y%m%d_%H_%M_%S")
    )
    shutil.copy2(
        "LOGS" + os.sep + scriptname + ".log",
        "LOGS" + os.sep + scriptname + "_" + ts + ".log",
    )

log = setup_logger(
    name=scriptname.lower(),
    log_file="LOGS" + os.sep + scriptname + ".log",
    logformat="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
    datefmt="%m-%d %H:%M",
)
log = setup_logger(
    name=scriptname.lower(),
    log_file="stderr",
    logformat="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%m-%d %H:%M",
)

import MONSDA.Params as mp
import MONSDA.Utils as mu
import MONSDA.Workflows as mw


# CODE
def parseargs():
    parser = argparse.ArgumentParser(
        description="Modular Organizer of Nextflow and Snakemake driven hts Data Analysis"
    )
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        nargs="+",
        action="append",
        help="Configuration json to read and optional config for nextflow",
    )
    parser.add_argument(
        "-d", "--directory", type=str, default="", help="Working Directory"
    )
    parser.add_argument(
        "-u",
        "--use-conda",
        action="store_true",
        default=True,
        help="Should conda be used, default True",
    )
    parser.add_argument(
        "-l",
        "--unlock",
        action="store_true",
        help="If Snakemake directory is locked you can unlock before processing",
    )
    parser.add_argument(
        "-j",
        "--procs",
        type=int,
        default=1,
        help="Number of parallel processes to start MONSDA with, capped by MAXTHREADS in config!",
    )
    parser.add_argument(
        "--save",
        action="store_true",
        help="Do not actually run jobs, create corresponding text file containing CLI-calls and arguments for manual running instead",
    )
    parser.add_argument(
        "-s",
        "--skeleton",
        action="store_true",
        help="Just create the minimal directory hierarchy as needed",
    )
    parser.add_argument(
        "--snakemake",
        action="store_true",
        default=True,
        help="Wrap around Snakemake, default",
    )
    parser.add_argument(
        "--nextflow", action="store_true", default=False, help="Wrap around Nextflow"
    )
    parser.add_argument(
        "--clean",
        action="store_true",
        help="Cleanup workdir (Nextflow), append -n to see list of files to clean or -f to actually remove those files",
    )
    parser.add_argument(
        "--loglevel",
        type=str,
        default="INFO",
        choices=["WARNING", "ERROR", "INFO", "DEBUG"],
        help="Set log level",
    )
    parser.add_argument(
        "--version",
        action="store_true",
        help="Print version and exit",
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return parser.parse_known_args()


def run_snakemake(
    configfile,
    workdir,
    useconda,
    procs,
    skeleton,
    loglevel,
    save=None,
    unlock=None,
    optionalargs=None,
):
    try:
        logid = scriptname + ".run_snakemake: "
        config = load_configfile(configfile)
        subdir = "SubSnakes"
        mp.create_skeleton(subdir, skeleton)

        argslist = list()
        if useconda:
            argslist.append("--conda-frontend conda --software-deployment-method conda")
        else:
            log.warning(
                logid
                + "You are not making use of conda, be aware that this will most likely not work for the workflows provided here! To change append the --use-conda option to your commandline call. You can also speed up conda with the --conda-frontend mamba argument and preinstall all conda environments appending the --software-deployment-method conda and the --create-envs-only arguments and share conda environment locations across runs with the --conda-prefix argument."
            )
        if optionalargs and len(optionalargs) > 0:
            log.debug(logid + "OPTIONALARGS: " + str(optionalargs))
            argslist.extend(optionalargs)
            if "--profile" in optionalargs and "MONSDA/slurm" in optionalargs:
                mu.makeoutdir("LOGS/SLURM")

        threads = (
            min(int(config["MAXTHREADS"]), procs) if "MAXTHREADS" in config else procs
        )
        config["MAXTHREADS"] = threads

        if unlock:
            try:
                pythonversion = (
                    f"python{str(sys.version_info.major)}.{str(sys.version_info.minor)}"
                )
                installpath = os.path.dirname(__file__).replace(
                    os.sep.join(["lib", pythonversion, "site-packages", "MONSDA"]),
                    "share",
                )
            except:
                installpath = os.getcwd()

            workflowpath = os.path.join(installpath, "MONSDA", "workflows")
            log.info(logid + "Unlocking directory")
            snk = os.path.abspath(os.path.join(workflowpath, "unlock.smk"))
            jobtorun = (
                f"snakemake --unlock -j {threads} -s {snk} --configfile {configfile}"
            )

            log.info(logid + "UNLOCKING " + str(jobtorun))
            job = runjob(jobtorun)
            log.debug(logid + "JOB CODE " + str(job))

        # Get processes to work on
        preprocess, subworkflows, postprocess = mw.get_processes(config)
        conditions = mp.get_conditions(config)

        """
        START TO PREPROCESS
        IF WE NEED TO DOWNLOAD FILES WE DO THIS NOW
        """
        if preprocess:
            for proc in [
                x for x in preprocess if config.get(x) and x in ["FETCH", "BASECALL"]
            ]:
                log.info(logid + "Preprocess " + str(proc))
                if proc not in config:
                    log.error(
                        logid
                        + "No configuration with key "
                        + proc
                        + " for file download found. Nothing to do!"
                    )
                mu.makeoutdir("FASTQ")
                mu.makeoutdir("TMP")

                if proc == "FETCH":
                    SAMPLES = mp.download_samples(config)
                    preprocess.remove(proc)
                elif proc == "BASECALL":
                    SAMPLES = mp.basecall_samples(config)
                    preprocess.remove(proc)
                else:
                    continue  # We only want download/basecall here

                log.debug(logid + "PRESAMPLES: " + str(SAMPLES))
                combinations = (
                    mw.get_combo(preprocess, config, conditions)
                    if subworkflows
                    else None
                )

                subwork = proc

                jobs = mw.make_pre(
                    subwork,
                    config,
                    SAMPLES,
                    conditions,
                    subdir,
                    loglevel,
                    "Pre",
                    combinations=combinations,
                )

                jobstorun = list()
                for job in jobs:
                    smko, confo = job
                    rest = " ".join(argslist)
                    jobstorun.append(
                        f"snakemake -j {threads} -s {smko} --configfile {confo} --directory {workdir} --printshellcmds --show-failed-logs {rest}"
                    )

                for job in jobstorun:
                    with open("JOBS" + os.sep + scriptname + ".commands", "a") as j:
                        j.write(job + os.linesep)
                    if not save:
                        log.info(logid + "RUNNING " + str(job))
                        jid = runjob(job)
                        log.debug(logid + "JOB CODE " + str(jid))

        """
        ONCE FILES ARE DOWNLOADED WE CAN START OTHER PREPROCESSING STEPS
        """
        mu.makeoutdir("TMP")
        SAMPLES = mp.get_samples(config)
        log.info(logid + "SAMPLES: " + str(SAMPLES))

        if preprocess:
            log.info(logid + "STARTING PREPROCESSING")
            if "QC" in preprocess and "QC" in config:
                mu.makeoutdir("QC")

                for subwork in preprocess:
                    combinations = (
                        mw.get_combo(subworkflows, config, conditions)
                        if subworkflows
                        else None
                    )
                    jobs = mw.make_pre(
                        subwork,
                        config,
                        SAMPLES,
                        conditions,
                        subdir,
                        loglevel,
                        "Pre",
                        combinations=combinations,
                    )

                    jobstorun = list()

                    for job in jobs:
                        smko, confo = job
                        rest = " ".join(argslist)

                        jobstorun.append(
                            f"snakemake -j {threads} --software-deployment-method conda -s {smko} --configfile {confo} --directory {workdir} --printshellcmds --show-failed-logs {rest}"
                        )

                    for job in jobstorun:
                        with open("JOBS" + os.sep + scriptname + ".commands", "a") as j:
                            j.write(job + os.linesep)
                        if not save:
                            log.info(logid + "RUNNING " + str(job))
                            jid = runjob(job)
                            log.debug(logid + "JOB CODE " + str(jid))

        else:
            log.warning(
                logid + "No preprocessing workflows defined! Continuing with workflows!"
            )

        """
        END OF PREPROCESSING, START OF PROCESSING
        """
        mu.makeoutdir("TMP")
        if subworkflows:
            combinations = (
                mw.get_combo(subworkflows, config, conditions) if subworkflows else None
            )
            jobs = mw.make_sub(
                subworkflows,
                config,
                SAMPLES,
                conditions,
                subdir,
                loglevel,
                combinations=combinations,
            )

            jobstorun = list()

            for job in jobs:
                smko, confo = job
                rest = " ".join(argslist)

                jobstorun.append(
                    f"snakemake -j {threads} --software-deployment-method conda -s {smko} --configfile {confo} --directory {workdir} --printshellcmds --show-failed-logs {rest}"
                )

            for job in jobstorun:
                with open("JOBS" + os.sep + scriptname + ".commands", "a") as j:
                    j.write(job + os.linesep)
                    if not save:
                        log.info(logid + "RUNNING " + str(job))
                        jid = runjob(job)
                        log.debug(logid + "JOB CODE " + str(jid))

        else:
            log.warning(
                logid
                + "No Workflows defined! Nothing to do, continuing with postprocessing!"
            )

        """
        END OF PROCESSING, START OF POSTPROCESSING
        """
        mu.makeoutdir("TMP")
        if postprocess:
            for subwork in postprocess:
                SAMPLES = mp.get_samples_postprocess(config, subwork)
                log.info(logid + f"POSTPROCESSING {subwork} with SAMPLES: {SAMPLES}")
                combinations = (
                    mw.get_combo(subworkflows, config, conditions)
                    if subworkflows
                    else None
                )

                jobs = mw.make_post(
                    subwork,
                    config,
                    SAMPLES,
                    conditions,
                    subdir,
                    loglevel,
                    combinations=combinations,
                )
                jobstorun = list()

                for job in jobs:
                    smko, confo = job
                    rest = " ".join(argslist)
                    jobstorun.append(
                        f"snakemake -j {threads} --software-deployment-method conda -s {smko} --configfile {confo} --directory {workdir} --printshellcmds --show-failed-logs {rest}"
                    )

                for job in jobstorun:
                    with open("JOBS" + os.sep + scriptname + ".commands", "a") as j:
                        j.write(job + os.linesep)
                        if not save:
                            log.info(logid + "RUNNING " + str(job))
                            jid = runjob(job)
                            log.debug(logid + "JOB CODE " + str(jid))

            if (
                any([x in postprocess for x in ["DE", "DEU", "DAS", "DTU"]])
                and not save
            ):
                # SUMMARY RUN
                combinations = (
                    mw.get_combo(subworkflows, config, conditions)
                    if subworkflows
                    else None
                )
                jobs = mw.make_summary(
                    config, subdir, loglevel, combinations=combinations
                )
                jobstorun = list()

                for job in jobs:
                    smko, confo = job
                    rest = " ".join(argslist)
                    jobstorun.append(
                        f"snakemake -j {threads} --software-deployment-method conda -s {smko} --configfile {confo} --directory {workdir} --printshellcmds --show-failed-logs {rest}"
                    )

                for job in jobstorun:
                    with open("JOBS" + os.sep + scriptname + ".commands", "a") as j:
                        j.write(job + os.linesep)
                        if not save:
                            log.info(logid + "RUNNING " + str(job))
                            jid = runjob(job)
                            log.debug(logid + "JOB CODE " + str(jid))

        else:
            log.warning(logid + "No postprocessing steps defined! Nothing to do!")

        if save:
            log.info(
                f"{logid} All CLI calls have been saved to monsda.commands in directory JOBS"
            )

    except Exception:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        log.error("".join(tbe.format()))


def run_nextflow(
    configfile,
    workdir,
    useconda,
    procs,
    skeleton,
    loglevel,
    save=None,
    clean=None,
    optionalargs=None,
):
    try:
        logid = scriptname + ".run_nextflow: "
        argslist = list()
        if useconda:
            argslist.append("-with-conda")
        else:
            log.warning(
                logid
                + "You are not making use of conda, be aware that this will most likely not work for the workflows provided here! To change append the --software-deployment-method conda option to your commandline call."
            )
        if optionalargs and len(optionalargs) > 0:
            log.debug(logid + "OPTIONALARGS: " + str(optionalargs))
            argslist.extend(optionalargs)

        if clean:
            log.info(logid + "Cleaning working directory")
            a = " ".join(argslist)
            jobtorun = f"nextflow clean {a}"
            log.info(logid + "CLEANUP " + str(jobtorun))
            job = runjob(jobtorun)
            log.debug(logid + "JOB CODE " + str(job))
            sys.exit()

        config = load_configfile(configfile)
        workdir = os.path.abspath(str.join(os.sep, [workdir, "NextFlowWork"]))
        subdir = "SubFlows"
        mp.create_skeleton(subdir, skeleton)

        if not os.path.exists(os.path.abspath(subdir + os.sep + "bin")):
            os.symlink(
                os.path.abspath(config.get("BINS", "")),
                os.path.abspath(subdir + os.sep + "bin"),
            )

        threads = (
            min(int(config["MAXTHREADS"]), procs) if "MAXTHREADS" in config else procs
        )
        config["MAXTHREADS"] = threads

        # Get processes to work on
        preprocess, subworkflows, postprocess = mw.nf_get_processes(config)
        conditions = mp.get_conditions(config)

        """
        START TO PROCESS
        IF WE NEED TO DOWNLOAD FILES WE DO THIS NOW
        """

        if preprocess:
            for proc in [
                x for x in preprocess if config.get(x) and x in ["FETCH", "BASECALL"]
            ]:
                log.info(logid + "Preprocess " + str(proc))
                if proc not in config:
                    log.error(
                        logid
                        + "No configuration with key "
                        + proc
                        + " for file download found. Nothing to do!"
                    )
                mu.makeoutdir("FASTQ")
                mu.makeoutdir("TMP")

                if proc == "FETCH":
                    if mp.check_samples(config) is True:
                        preprocess.remove(proc)
                        SAMPLES = None
                    else:
                        SAMPLES = mp.download_samples(config)
                        preprocess.remove(proc)
                elif proc == "BASECALL":
                    SAMPLES = mp.basecall_samples(config)
                    preprocess.remove(proc)
                else:
                    continue
                if SAMPLES is None:
                    log.info(
                        logid + f"All SAMPLES already available, skipping FETCH process"
                    )
                    continue

                log.debug(logid + "PRESAMPLES: " + str(SAMPLES))
                combinations = (
                    mw.get_combo(preprocess, config, conditions)
                    if subworkflows
                    else None
                )

                subwork = proc

                jobs = mw.nf_make_pre(
                    subwork,
                    config,
                    SAMPLES,
                    conditions,
                    subdir,
                    loglevel,
                    "Pre",
                    combinations=combinations,
                )

                jobstorun = list()
                for job in jobs:
                    nfo, confo, tp, params = job
                    pars = " ".join(
                        "--{!s} {!s}".format(key, val) for (key, val) in params.items()
                    )
                    rest = " ".join(argslist)

                    jobstorun.append(
                        f"nextflow -log {workdir}/run.log run {nfo} -w {workdir} {rest} {pars} {tp}"
                    )

                for job in jobstorun:
                    with open("JOBS" + os.sep + scriptname + ".commands", "a") as j:
                        j.write(job + os.linesep)
                    if not save:
                        log.info(logid + "RUNNING " + str(job))
                        jid = runjob(job)
                        log.debug(logid + "JOB CODE " + str(jid))

        """
        ONCE FILES ARE DOWNLOAD WE CAN START OTHER PREPROCESSING STEPS
        """
        mu.makeoutdir("TMP")
        SAMPLES = mp.get_samples(config)
        log.info(logid + "SAMPLES: " + str(SAMPLES))
        conditions = mp.get_conditions(config)
        log.info(logid + "CONDITIONS: " + str(conditions))

        if preprocess:
            log.info(logid + "STARTING PREPROCESSING")
            if "QC" in preprocess and "QC" in config:
                mu.makeoutdir("QC")

            for subwork in preprocess:
                combinations = (
                    mw.get_combo(subworkflows, config, conditions)
                    if subworkflows
                    else None
                )
                jobs = mw.nf_make_pre(
                    subwork,
                    config,
                    SAMPLES,
                    conditions,
                    subdir,
                    loglevel,
                    combinations=combinations,
                )

                jobstorun = list()

                for job in jobs:
                    nfo, confo, tp, params = job
                    pars = " ".join(
                        "--{!s} {!s}".format(key, val) for (key, val) in params.items()
                    )
                    rest = " ".join(argslist)

                    jobstorun.append(
                        f"nextflow -log {workdir}/run.log run {nfo} -w {workdir} {rest} {pars} {tp}"
                    )
                    log.debug(f"{logid} JOBS: {jobstorun}")

                for job in jobstorun:
                    with open("JOBS" + os.sep + scriptname + ".commands", "a") as j:
                        j.write(job + os.linesep)
                    if not save:
                        log.info(logid + "RUNNING " + str(job))
                        jid = runjob(job)
                        log.debug(logid + "JOB CODE " + str(jid))

        else:
            log.warning(
                logid + "No preprocessing workflows defined! Continuing with workflows!"
            )

        """
        END OF PREPROCESSING, START OF PROCESSING
        """
        mu.makeoutdir("TMP")
        if subworkflows:
            combinations = (
                mw.get_combo(subworkflows, config, conditions) if subworkflows else None
            )
            jobs = mw.nf_make_sub(
                subworkflows,
                config,
                SAMPLES,
                conditions,
                subdir,
                loglevel,
                combinations=combinations,
            )

            jobstorun = list()

            for job in jobs:
                nfo, confo, tp, params = job
                pars = " ".join(
                    "--{!s} {!s}".format(key, val) for (key, val) in params.items()
                )
                rest = " ".join(argslist)

                jobstorun.append(
                    f"nextflow -log {workdir}/run.log run {nfo} -w {workdir} {rest} {pars} {tp}"
                )

            log.debug(f"{logid} JOBS: {jobstorun}")
            for job in jobstorun:
                with open("JOBS" + os.sep + scriptname + ".commands", "a") as j:
                    j.write(job + os.linesep)
                    if not save:
                        log.info(logid + "RUNNING " + str(job))
                        jid = runjob(job)
                        log.debug(logid + "JOB CODE " + str(jid))

        else:
            log.warning(
                logid
                + "No Workflows defined! Nothing to do, continuing with postprocessing!"
            )

        """
        END OF PROCESSING, START OF POSTPROCESSING
        """
        mu.makeoutdir("TMP")
        if postprocess:
            for subwork in postprocess:
                SAMPLES = mp.get_samples_postprocess(config, subwork)
                log.info(logid + "POSTPROCESSING SAMPLES: " + str(SAMPLES))
                combinations = (
                    mw.get_combo(subworkflows, config, conditions)
                    if subworkflows
                    else None
                )
                log.debug(logid + "POSTPROCESSING WITH COMBOS: " + str(combinations))

                jobs = mw.nf_make_post(
                    subwork,
                    config,
                    SAMPLES,
                    conditions,
                    subdir,
                    loglevel,
                    combinations=combinations,
                )
                jobstorun = list()

                for job in jobs:
                    nfo, confo, tp, params = job
                    pars = " ".join(
                        "--{!s} {!s}".format(key, val) for (key, val) in params.items()
                    )
                    rest = " ".join(argslist)

                    jobstorun.append(
                        f"nextflow -log {workdir}/run.log run {nfo} -w {workdir} {rest} {pars} {tp}"
                    )

                for job in jobstorun:
                    with open("JOBS" + os.sep + scriptname + ".commands", "a") as j:
                        j.write(job + os.linesep)
                        if not save:
                            log.info(logid + "RUNNING " + str(job))
                            jid = runjob(job)
                            log.debug(logid + "JOB CODE " + str(jid))

            if (
                any([x in postprocess for x in ["DE", "DEU", "DAS", "DTU"]])
                and not save
            ):
                # SUMMARY RUN
                combinations = (
                    mw.get_combo(subworkflows, config, conditions)
                    if subworkflows
                    else None
                )
                jobs = mw.nf_make_summary(
                    config, subdir, loglevel, combinations=combinations
                )
                jobstorun = list()

                for job in jobs:
                    nfo, confo, tp, params = job
                    pars = " ".join(
                        "--{!s} {!s}".format(key, val) for (key, val) in params.items()
                    )
                    rest = " ".join(argslist)

                    jobstorun.append(
                        f"nextflow -log {workdir}/run.log run {nfo} -w {workdir} {rest} {pars} {tp}"
                    )

                for job in jobstorun:
                    with open("JOBS" + os.sep + scriptname + ".commands", "a") as j:
                        j.write(job + os.linesep)
                        if not save:
                            log.info(logid + "RUNNING " + str(job))
                            jid = runjob(job)
                            log.debug(logid + "JOB CODE " + str(jid))

        else:
            log.warning(logid + "No postprocessing steps defined! Nothing to do!")

        if save:
            log.info(
                f"{logid} All CLI calls have been saved to monsda.commands in directory JOBS"
            )

    except Exception:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        log.error("".join(tbe.format()))


def runjob(jobtorun):
    try:
        logid = scriptname + ".runjob: "
        # return subprocess.run(jobtorun, shell=True, universal_newlines=True, capture_output=True)  # python >= 3.7
        job = subprocess.Popen(
            jobtorun,
            shell=True,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=1,
            close_fds=True,
        )

        while True:
            output, outerr = job.communicate()
            output = str.join("", output).rstrip()
            outerr = str.join("", outerr).rstrip()
            if output == "" and outerr == "" and job.poll() is not None:
                break
            if output and output != "":
                if (
                    any(x in output for x in ["ERROR", "Error", "error", "Exception"])
                    and not "Workflow finished" in output
                    or "Workflow finished" in outerr
                ):
                    if outerr:
                        log.error(
                            logid + "STOPPING: " + str(output) + "\n" + str(outerr)
                        )
                    else:
                        log.error(logid + "STOPPING: " + str(output))
                    log.info("PLEASE CHECK LOG AT LOGS/MONSDA.log")
                    job.kill()
                    sys.exit()
                else:
                    log.info(logid + str(output))
            if outerr and outerr != "":
                if (
                    not "Workflow finished" in outerr
                    and not "Nothing to be done" in outerr
                    and not "Workflow finished" in output
                    and any(
                        x in outerr for x in ["ERROR", "Error", "error", "Exception"]
                    )
                ):
                    log.error(logid + "STOPPING: " + str(outerr))
                    log.info("PLEASE CHECK LOG AT LOGS/MONSDA.log")
                    job.kill()
                    sys.exit()
                else:
                    log.info(logid + str(outerr))
            if job.poll() is not None:
                break

        if job.returncode == 0:
            output, outerr = job.communicate()
            output = str.join("", output).rstrip()
            if output and output != "":
                log.info(logid + "JOB FINISHED: " + output)
            return job.poll()
        else:
            output, outerr = job.communicate()
            output = str.join("", output).rstrip()
            outerr = str.join("", outerr).rstrip()
            if outerr and outerr != "" or output and output != "":
                log.error(logid + "ERROR: " + outerr + output)
                log.info("PLEASE CHECK LOG AT LOGS/MONSDA.log")
            job.kill()
            sys.exit("ERROR SIGNAL: " + str(job.returncode))

    except Exception:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        log.error("".join(tbe.format()))
        sys.exit()


def main():
    logid = scriptname + ".main: "
    try:
        args = parseargs()
        knownargs = args[0]
        optionalargs = args[1:]

        if knownargs.version:
            sys.exit("MONSDA version " + __version__)

        log.setLevel(knownargs.loglevel)

        knownargs.configfile = str(knownargs.config[0].pop())

        if len(knownargs.config) > 1:
            for i in range(1, len(knownargs.config)):
                optionalargs[0].extend(list(["-c", str(knownargs.config[i].pop())]))

        log.debug(
            f"{logid} ARGS: {args} {type(args)} KNOWNARGS: {knownargs} {type(knownargs)} OPTIONALARGS: {optionalargs} {type(optionalargs)}"
        )

        log.info(
            logid
            + "MONSDA-CLI: "
            + sys.argv[0]
            + " "
            + "{}".format(" ".join([shlex.quote(s) for s in sys.argv[1:]]))
        )

        required_version = load_configfile(knownargs.configfile).get("VERSION")
        if not required_version:
            sys.exit(
                "Can not check version needed, please add VERSION key to config file"
            )

        if mw.monsda_check_version(__version__, required_version) is not True:
            log.error(
                "Version required in config file "
                + str(required_version)
                + " does not match installed version "
                + str(__version__)
                + " Please install correct version before continuing!"
            )
            sys.exit(
                "Version required in config file "
                + str(required_version)
                + " does not match installed version "
                + str(__version__)
                + " Please install correct version before continuing!"
            )
        else:
            log.info("Running MONSDA version " + __version__ + " as configured")

        MIN_PYTHON = (3, 11)
        if sys.version_info < MIN_PYTHON:
            log.error("This script requires Python version >= 3.7")
            sys.exit("This script requires Python version >= 3.7")
        log.info(
            logid + "Running " + scriptname + " on " + str(knownargs.procs) + " cores"
        )
        log.debug(logid + str(log.handlers))

        commandfile = "JOBS" + os.sep + scriptname + ".commands"
        if os.path.exists(commandfile):
            ts = "_" + str(
                datetime.datetime.fromtimestamp(
                    os.path.getmtime(os.path.abspath(commandfile))
                ).strftime("%Y%m%d_%H_%M_%S")
            )
            shutil.copy2(
                commandfile,
                commandfile + ts,
            )

        if not knownargs.nextflow:
            mw.sm_check_version("8.10.0")
            run_snakemake(
                knownargs.configfile,
                knownargs.directory,
                knownargs.use_conda,
                knownargs.procs,
                knownargs.skeleton,
                knownargs.loglevel,
                knownargs.save,
                knownargs.unlock,
                optionalargs[0],
            )

        else:
            nf_min_version = "23.10.1"
            nf_ver = mw.nf_check_version(nf_min_version)
            if nf_ver:
                run_nextflow(
                    knownargs.configfile,
                    knownargs.directory,
                    knownargs.use_conda,
                    knownargs.procs,
                    knownargs.skeleton,
                    knownargs.loglevel,
                    knownargs.save,
                    knownargs.clean,
                    optionalargs[0],
                )
            else:
                log.error(
                    logid
                    + "Minimal version of nextflow required is "
                    + str(nf_min_version)
                    + " and we only found "
                    + str(nf_ver)
                    + "! Please install or use envs/MONSDA.yaml to create conda environment accordingly"
                )
    except Exception:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        log.error(logid + "".join(tbe.format()))


####################
####    MAIN    ####
####################


if __name__ == "__main__":
    logid = scriptname + ".main: "
    try:
        main()

    except Exception:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        log.error(logid + "".join(tbe.format()))

# MONSDA.py ends here



# Params.py ---
#
# Filename: Params.py
# Description:
# Author: Joerg Fallmann
# Maintainer:
# Created: Tue Sep 18 15:39:06 2018 (+0200)
# Version:
# Package-Requires: ()
# Last-Updated: Thu Feb  4 18:01:07 2021 (+0100)
#           By: Joerg Fallmann
#     Update #: 2888
# URL:
# Doc URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#

# Change Log:
#
#
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
#
#

# Code:
# import os, sys, inspect
# # realpath() will make your script run, even if you symlink it :)
# cmd_folder = os.path.dirname(os.path.realpath(os.path.abspath( inspect.getfile( inspect.currentframe() )) ))
# if cmd_folder not in sys.path:
#     sys.path.insert(0, cmd_folder)

#
# # Use this if you want to include modules from a subfolder
# cmd_subfolder = os.path.join(os.path.dirname(os.path.realpath(os.path.abspath( inspect.getfile( inspect.currentframe() )) )),"lib")
# if cmd_subfolder not in sys.path:
#     sys.path.insert(0, cmd_subfolder)
#
# # Info:
# # cmd_folder = os.path.dirname(os.path.abspath(__file__)) # DO NOT USE __file__ !!!
# # __file__ fails if the script is called in different ways on Windows.
# # __file__ fails if someone does os.chdir() before.
# # sys.argv[0] also fails, because it doesn't not always contain the path.

import collections
import datetime
import functools
import glob
import gzip
import hashlib
import heapq
import inspect
import itertools
import json
import logging
import os
import re
import shutil
import subprocess
import sys
import traceback as tb
from collections import OrderedDict, defaultdict
from io import StringIO
from operator import itemgetter

import numpy as np
import six
from Bio import SeqIO
from natsort import natsorted
from snakemake.common.configfile import load_configfile

import MONSDA.Utils as mu
from MONSDA.Utils import check_run as check_run

# cmd_subfolder = [os.path.join(os.path.dirname(os.path.realpath(os.path.abspath(inspect.getfile( inspect.currentframe() )) )),"../MONSDA/lib"), os.path.join(os.path.dirname(os.path.realpath(os.path.abspath(inspect.getfile( inspect.currentframe() )) )),"MONSDA/lib"), os.path.join(os.path.dirname(os.path.realpath(os.path.abspath(inspect.getfile( inspect.currentframe() )) )),"../lib"), os.path.join(os.path.dirname(os.path.realpath(os.path.abspath(inspect.getfile( inspect.currentframe() )) )),"lib")]
# for x in cmd_subfolder:
# if x not in sys.path:
#        sys.path.insert(0, x)


try:
    scriptname = (
        os.path.basename(inspect.stack()[-1].filename)
        .replace("Run", "")
        .replace(".py", "")
    )
    log = logging.getLogger(scriptname)

    lvl = log.level if log.level else "INFO"
    for handler in log.handlers[:]:
        handler.close()
        log.removeHandler(handler)
    handler = logging.FileHandler("LOGS/MONSDA.log", mode="a")
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    log.setLevel(lvl)

except Exception:
    exc_type, exc_value, exc_tb = sys.exc_info()
    tbe = tb.TracebackException(
        exc_type,
        exc_value,
        exc_tb,
    )
    print("".join(tbe.format()), file=sys.stderr)


@check_run
def get_samples(config):
    logid = scriptname + ".Params_get_samples: "
    SAMPLES = [os.path.join(x) for x in sampleslong(config)]
    log.debug(logid + "SAMPLES_LONG: " + str(SAMPLES))
    check = [
        os.path.join("FASTQ", str(x).replace(".fastq.gz", "") + "*.fastq.gz")
        for x in SAMPLES
    ]
    RETSAMPLES = list()
    for i in range(len(check)):
        s = check[i]
        log.debug(logid + "SEARCHING: " + s)
        paired = checkpaired([SAMPLES[i]], config)
        log.debug(logid + "PAIRED: " + str(paired))
        f = glob.glob(s)
        log.debug(logid + "SAMPLECHECK: " + str(f))
        if f:
            f = list(set([str.join(os.sep, s.split(os.sep)[1:]) for s in f]))
            if "paired" not in paired:
                RETSAMPLES.extend(
                    list(
                        set(
                            [
                                os.path.join(
                                    os.path.dirname(s),
                                    re.sub(
                                        r"_r1.fastq.gz|_R1.fastq.gz|_r2.fastq.gz|_R2.fastq.gz|.fastq.gz",
                                        "",
                                        os.path.basename(s),
                                    ),
                                )
                                for s in f
                            ]
                        )
                    )
                )
                log.debug(logid + "PAIREDSAMPLES: " + str(f))
            else:
                RETSAMPLES.extend([x.replace(".fastq.gz", "") for x in f])
        else:
            log.debug(logid + "SAMPLECHECK: " + str(f))
    log.debug(logid + "SAMPLETEST: " + str(RETSAMPLES))
    if len(RETSAMPLES) < 1:
        log.error(logid + "No samples found, please check config file")
        sys.exit()

    log.debug(logid + "SAMPLES: " + str(RETSAMPLES))
    return RETSAMPLES


@check_run
def get_samples_postprocess(config, subwork):
    logid = scriptname + ".Params_get_samples_postprocess: "
    SAMPLES = [
        os.path.join(x)
        for x in sampleslong(config)
        if len(mu.get_from_dict(config[subwork], conditiononly(x, config))) > 0
        and (
            not config[subwork].get("EXCLUDE")
            or os.path.basename(x) not in config[subwork]["EXCLUDE"]
        )
    ]  # only those samples where postprocessing steps are defined for in config, should we make this a per condition check?
    log.debug(logid + "SAMPLES_LONG: " + str(SAMPLES))
    check = [
        os.path.join("FASTQ", str(x).replace(".fastq.gz", "") + "*.fastq.gz")
        for x in SAMPLES
    ]
    RETSAMPLES = list()
    for i in range(len(check)):
        s = check[i]
        paired = checkpaired([SAMPLES[i]], config)
        log.debug(logid + "PAIRED: " + str(paired))
        f = glob.glob(s)
        log.debug(logid + "SAMPLECHECK: " + str(f))
        if f:
            f = sorted(list(set([str.join(os.sep, s.split(os.sep)[1:]) for s in f])))
            if "paired" in paired:
                RETSAMPLES.extend(
                    sorted(
                        list(
                            set(
                                [
                                    os.path.join(
                                        os.path.dirname(s),
                                        re.sub(
                                            r"_r1.fastq.gz|_R1.fastq.gz|_r2.fastq.gz|_R2.fastq.gz|.fastq.gz",
                                            "",
                                            os.path.basename(s),
                                        ),
                                    )
                                    for s in f
                                ]
                            )
                        )
                    )
                )
                log.debug(logid + "PAIREDSAMPLES: " + str(f))
            else:
                RETSAMPLES.extend([x.replace(".fastq.gz", "") for x in f])
        else:
            log.debug(logid + "SAMPLECHECK: " + str(f))
    log.debug(logid + "SAMPLETEST: " + str(RETSAMPLES))
    if len(RETSAMPLES) < 1:
        log.error(
            logid
            + "No samples found for "
            + str(subwork)
            + ", please check config file"
        )
        sys.exit()

    log.debug(logid + "SAMPLES: " + str(RETSAMPLES))
    return RETSAMPLES


@check_run
def check_samples(config):
    logid = scriptname + ".Params_check_samples: "
    tocheck = [os.path.join(x) for x in sampleslong(config, nocheck=True)]
    log.debug(logid + "SEARCHING: " + str(tocheck))
    check = [
        os.path.join("FASTQ", str(x).replace(".fastq.gz", "") + "*.fastq.gz")
        for x in tocheck
    ]
    for i in range(len(check)):
        s = check[i]
        log.debug(logid + "SEARCHING: " + s)
        f = glob.glob(s)
        log.debug(logid + "SAMPLECHECK: " + str(f))
        if f:
            continue
        else:
            return False
    return True


@check_run
def download_samples(config):
    logid = scriptname + ".Params_download_samples: "
    SAMPLES = [os.path.join(x) for x in sampleslong(config, nocheck=True)]
    log.debug(logid + "DOWNLOAD_SAMPLES_LONG: " + str(SAMPLES))
    return SAMPLES


@check_run
def basecall_samples(config):
    logid = scriptname + ".Params_basecall_samples: "
    SAMPLES = [os.path.join(x) for x in sampleslong(config, nocheck=True)]
    log.debug(logid + "SAMPLES_LONG: " + str(SAMPLES))
    check = [
        os.path.join("RAW", str(x).replace(".fast5", "") + "*.fast5") for x in SAMPLES
    ]
    check.extend(
        [os.path.join("RAW", str(x).replace(".pod5", "") + "*.pod5") for x in SAMPLES]
    )
    RETSAMPLES = list()
    for i in range(len(check)):
        s = check[i]
        log.debug(logid + "SEARCHING: " + s)
        f = glob.glob(s)
        log.debug(logid + "SAMPLECHECK: " + str(f))
        if f:
            f = list(set([str.join(os.sep, s.split(os.sep)[1:]) for s in f]))
            RETSAMPLES.extend([x.replace(".fast5", "").replace(".pod5", "") for x in f])
    log.debug(logid + "SAMPLETEST: " + str(RETSAMPLES))
    if len(RETSAMPLES) < 1:
        log.error(logid + "No samples found, please check config file")
        sys.exit()

    log.debug(logid + "SAMPLES: " + str(RETSAMPLES))
    return list(set(RETSAMPLES))


@check_run
def get_conditions(config, stage="SETTINGS"):
    logid = scriptname + ".Params_conditions: "
    ret = list()
    for k in mu.keysets_from_dict(config[stage], "SAMPLES"):
        ret.append(k)
    log.debug(logid + str(ret))
    return sorted(list(set(ret)))


@check_run
def get_samples_from_dir(search, config, nocheck=None):  # CHECK
    logid = scriptname + ".Params_get_samples_from_dir: "
    samples = [
        x.replace(" ", "")
        for x in list(set(mu.get_from_dict(config["SETTINGS"], search)[0]["SAMPLES"]))
    ]
    log.debug(logid + f"Samples: {samples}, Search: {search}, Check: {nocheck}")
    if nocheck is not None:
        samples = [os.sep.join([os.sep.join(search[0:]), s]) for s in samples]
        return list(set(samples))
    for x in range(
        len(search), len(search) - 2, -1
    ):  # For arbitrary depth of ics we append subdirectories until samples are found, maximum of one setting additional to file path is allowed
        pat = os.sep.join(["FASTQ", os.sep.join(search[0:x]), "*.fastq.gz"])
        log.debug(logid + "REGEX: " + str(pat) + "\t" + "SAMPLES: " + str(samples))
        check = natsorted(glob.glob(pat), key=lambda y: y.lower())
        log.debug(logid + "check: " + str(check))
        if len(check) > 0:
            ret = list()
            clean = list()
            for (
                c
            ) in (
                check
            ):  # If sample fits glob pattern but is not actually in the part of the config we are looking at this needs to be checked as checkpaired returns None otherwise, e.g. Condition1 has Sample abc_R1.fastq and Condition2 has Sample abcd_R1.fastq
                log.debug(logid + "check: " + str(c))
                x = c.split(os.sep)[-1]
                for s in samples:
                    log.debug(logid + "x: " + str(x))
                    log.debug(logid + "sample: " + str(s))
                    if re.match(f"^{s}_R", x) or x == s + ".fastq.gz":
                        log.debug(
                            logid
                            + "FOUND: "
                            + s
                            + "_R"
                            + " or "
                            + s
                            + ".fastq.gz"
                            + " in "
                            + x
                        )
                        clean.append(c)
                        break
            log.debug(logid + "checkclean: " + str(clean))
            paired = checkpaired(
                [os.sep.join([os.sep.join(search), clean[0].split(os.sep)[-1]])], config
            )
            if paired is not None and "paired" in paired or "singlecell" in paired:
                log.debug(
                    logid
                    + "SEARCHING: "
                    + str(
                        [
                            os.sep.join(
                                [
                                    os.sep.join(os.path.dirname(s).split(os.sep)[1:]),
                                    re.sub(
                                        r"_r1.fastq.gz|_R1.fastq.gz|_r2.fastq.gz|_R2.fastq.gz|.fastq.gz",
                                        "",
                                        os.path.basename(s),
                                    ),
                                ]
                            )
                            for s in clean
                        ]
                    )
                )
                ret.extend(
                    list(
                        set(
                            [
                                os.sep.join(
                                    [
                                        os.sep.join(
                                            os.path.dirname(s).split(os.sep)[1:]
                                        ),
                                        re.sub(
                                            r"_r1.fastq.gz|_R1.fastq.gz|_r2.fastq.gz|_R2.fastq.gz|.fastq.gz",
                                            "",
                                            os.path.basename(s),
                                        ),
                                    ]
                                )
                                for s in clean
                            ]
                        )
                    )
                )
                log.debug(logid + "FOUND: " + str(ret))
                renamelist = [
                    re.sub(r"_r\d", lambda pat: pat.group(1).upper(), s) for s in ret
                ]
                for i in range(len(renamelist)):
                    if renamelist[i] != ret[i]:
                        log.warning(
                            "SAMPLE NAMES CONTAIN LOWER CASE r1/r2 INSTEAD OF R1/R2 FOR PAIRED END SEQUENCING, THEY WILL BE RENAMED"
                        )
                        os.rename(ret[i], renamelist[i])
            else:
                log.debug(
                    logid
                    + "SEARCHING: "
                    + str(
                        [
                            os.sep.join(s.split(os.sep)[1:]).replace(".fastq.gz", "")
                            for s in clean
                        ]
                    )
                )
                ret.extend(
                    [
                        os.sep.join(s.split(os.sep)[1:]).replace(".fastq.gz", "")
                        for s in clean
                    ]
                )

            log.debug(logid + "RETURN: " + str(ret))
            return list(set(ret))
    log.error(logid + "NO SAMPLES FOUND")
    return list()


@check_run
def sampleslong(config, nocheck=None):
    logid = scriptname + ".Params_sampleslong: "
    log.debug(logid + f"Check: {nocheck}")
    tosearch = list()
    ret = list()
    for k in mu.keysets_from_dict(config["SETTINGS"], "SAMPLES"):
        tosearch.append(list(k))
    log.debug(logid + "tosearch: " + str(tosearch))
    for search in tosearch:
        ret.extend(get_samples_from_dir(search, config, nocheck))
    log.debug(logid + str(ret))
    return ret


@check_run
def get_placeholder(config):
    ret = list()
    if "PH" in (config):
        for x in config["PH"]:
            ret.append(str(x))
    else:
        ret.append("_")
    return ret


@check_run
def get_cutoff_as_string(config, subwork, cf):
    logid = scriptname + ".get_cutoff: "
    cutoff = (
        str(config[subwork]["CUTOFFS"].get(cf))
        if config[subwork].get("CUTOFFS")
        else ".05" if cf == "pval" else "1.5"
    )
    log.info(logid + "CUTOFFS: " + str(cf) + ":" + cutoff)
    return cutoff


@check_run
def get_summary_dirs(config):
    logid = scriptname + ".get_summary_dirs: "
    ret = list()
    for work, tools in config["WORKFLOWS"].items():
        for tool in tools:
            ret.append(f"{work}/{tool.upper()}")
    log.debug(logid + str(ret))
    return ret


@check_run
def get_summary_files(config):
    logid = scriptname + ".get_summary_files: "
    ret = list()
    for work, tools in config["WORKFLOWS"].items():
        for tool in tools:
            log.info(logid + "make summary of " + str(work) + " - " + str(tool))
            for f in glob.glob(f"{work}/{tool.upper()}/Sig*"):
                # for f in glob.glob(f"{work}/{tool.upper()}/*"):
                ret.append(f)
    log.debug(logid + str(ret))
    return ret


@check_run
def create_skeleton(runner, skeleton=None):
    logid = scriptname + ".Params_create_skeleton: "
    if skeleton:
        for subdir in ["SubSnakes", "RAW", "FASTQ", "LOGS", "TMP", "JOB"]:
            mu.makeoutdir(subdir)
            sys.exit(
                "Skeleton directories created, please add files and rerun without --skeleton option"
            )
    else:
        for subdir in [runner, "LOGS", "TMP", "JOBS"]:
            mu.makeoutdir(subdir)
        if os.path.isfile(os.path.abspath("JOBS" + os.sep + scriptname + ".commands")):
            ts = str(
                datetime.datetime.fromtimestamp(
                    os.path.getmtime(
                        os.path.abspath("JOBS" + os.sep + scriptname + ".commands")
                    )
                ).strftime("%Y%m%d_%H_%M_%S")
            )
            shutil.copy2(
                "JOBS" + os.sep + scriptname + ".commands",
                "JOBS" + os.sep + scriptname + "_" + ts + ".commands",
            )
            open("JOBS" + os.sep + scriptname + ".commands", "w").close()


@check_run
def tool_params(sample, runstate, config, subconf, tool=None):
    logid = scriptname + ".Params_tool_params: "
    log.debug(logid + "Samples: " + str(sample))
    mp = OrderedDict()
    x = sample.split(os.sep)[:-1]
    if runstate is None:
        runstate = runstate_from_sample([sample], config)[0]
    if runstate not in x:
        x.append(runstate)
    log.debug(logid + str([sample, runstate, subconf, x]))
    if "_" in tool:
        tool = tool.split("_")[0]
    mp = (
        mu.sub_dict(config[subconf], x)[tool]
        if tool
        else mu.sub_dict(config[subconf], x)
    )
    log.debug(logid + "DONE: " + str(mp))
    return mp


@check_run
def setting_per_sample(sample, runstate, config, setting, subconf=None):
    logid = scriptname + ".Params_setting_per_sample: "
    log.debug(logid + "Samples: " + str(sample))
    set = None
    x = sample.split(os.sep)[2:-1]
    if runstate is None:
        runstate = runstate_from_sample([sample], config)[0]
    if runstate not in x:
        x.append(runstate)
    subsetting = mu.sub_dict(config["SETTINGS"], x).get(setting)

    if setting == "ANNOTATION":  # Special case is annotation
        subsetting = subsetting.get(
            "GTF", subsetting.get("GFF")
        )  # by default GTF format will be used

    if subconf:  # check specific setting for workflow part
        subset = (
            config[subconf].get(setting)
            if config[subconf].get(setting)
            else mu.sub_dict(config[subconf], x).get(setting)
        )

    # Define which final setting is returned
    set = subset if subset else setting

    return set


@check_run
def get_reps(samples, config, analysis, process="smk"):
    logid = scriptname + ".Params_get_reps: "
    log.debug(logid + "Samples: " + str(samples))
    ret = defaultdict(list)
    for sample in samples:
        if process == "smk":
            checkid = sample.split(os.sep)[4]
            scond = (
                sample.split(os.sep)[4:-1]
                if config["SETTINGS"].get(checkid, False)
                else sample.split(os.sep)[3:-1]
            )
        else:
            scond = sample.split(os.sep)[:-1]
        log.debug(logid + "WORKING ON: " + str(sample) + " CONDITION: " + str(scond))
        partconf = mu.sub_dict(config["SETTINGS"], scond)
        log.debug(logid + "CONF: " + str(partconf))

        Ex = config[analysis].get("EXCLUDE")
        if Ex and sample.split(os.sep)[-1] in Ex:
            continue
        if process == "smk":
            ret["reps"].append(sample)
        else:
            ret["reps"].append(os.path.basename(sample))
        wcfile = sample.split(os.sep)[-1].replace("_mapped_sorted_unique.counts.gz", "")
        idx = partconf["SAMPLES"].index(wcfile)
        scond.append(sample.split(os.sep)[-1])
        ret["pairs"].append(checkpaired_rep([str.join(os.sep, scond)], config))
        ret["conds"].append(partconf["GROUPS"][idx])
        if (
            "BATCHES" in partconf
            and len(partconf["BATCHES"]) >= idx + 1
            and partconf.get("BATCHES") != ""
        ):
            ret["batches"].append(str(partconf["BATCHES"][idx]).replace(",", "_"))
        else:
            ret["batches"].append("1")
        if (
            "TYPES" in partconf
            and len(partconf["TYPES"]) >= idx + 1
            and partconf.get("TYPE") != ""
        ):
            ret["types"].append(str(partconf["TYPES"][idx]).replace(",", "_"))
        else:
            ret["types"].append("std")

    rets = "-r " + str.join(",", ret["reps"])
    rets += " -c " + str.join(",", ret["conds"])
    rets += " -t " + str.join(",", ret["types"])
    rets += " -b " + str.join(",", ret["batches"])
    rets += " --paired " + str.join(",", ret["pairs"]) if "pairs" in ret else ""

    log.debug(logid + "RETURN: " + str(rets))
    return rets


@check_run
def get_diego_samples(samples, config, analysis):
    logid = scriptname + ".Params_get_diego_samples: "
    log.debug(logid + "Samples: " + str(samples))
    ret = defaultdict(list)
    for sample in samples:
        scond = sample.split(os.sep)[4:-1]
        log.debug(logid + "WORKING ON: " + str(sample) + " CONDITION: " + str(scond))
        wcfile = str.join("-", sample.split(os.sep)[-4:]).replace(
            "_mapped_sorted_unique.counts.gz", ""
        )
        ret[wcfile].append(sample)

    log.debug(logid + "RETURN: " + str(ret))

    slist = ""
    for key, val in ret.items():
        slist += str(key) + "\t"
        slist += "\t".join(val)
        slist += os.linesep

        log.debug(logid + "RETURN: " + str(slist))
    return slist


@check_run
def get_diego_groups(samples, config, analysis):
    logid = scriptname + ".Params_get_diego_groups: "
    log.debug(logid + "Samples: " + str(samples))
    ret = defaultdict(list)
    for sample in samples:
        scond = sample.split(os.sep)[4:-1]
        log.debug(logid + "WORKING ON: " + str(sample) + " CONDITION: " + str(scond))
        partconf = mu.sub_dict(config["SETTINGS"], scond)
        log.debug(logid + "CONF: " + str(partconf))
        wcfile = str.join("-", sample.split(os.sep)[-4:]).replace(
            "_mapped_sorted_unique.counts.gz", ""
        )
        checkfile = sample.split(os.sep)[-1].replace(
            "_mapped_sorted_unique.counts.gz", ""
        )
        idx = partconf["SAMPLES"].index(checkfile)
        cond = partconf["GROUPS"][idx]
        ret[cond].append(wcfile)

    slist = ""
    for key, val in ret.items():
        slist += str(key) + "\t"
        slist += "\t".join(val)
        slist += os.linesep
    log.debug(logid + "RETURN: " + str(slist))
    return slist


@check_run
def env_bin_from_config2(samples, config, subconf):
    logid = scriptname + ".Params_env_bin_from_config2: "
    for s in samples:
        log.debug(logid + "S: " + str(s))
        log.debug(logid + "C: " + str(conditiononly(s, config)))
        check = conditiononly(s, config)

        for k in mu.get_from_dict(config[subconf], check):
            if "BIN" in k:
                mb = k["BIN"]
            else:
                mb = ""
            if "ENV" in k:
                me = k["ENV"]
            else:
                me = ""
        log.debug(logid + str([str(mb), str(me)]))
    return mb, me


@check_run
def env_bin_from_config(config, subconf):
    logid = scriptname + ".Params_env_bin_from_config: "
    envkey = subconf + "ENV"
    binkey = subconf + "BIN"
    me = config[envkey]
    mb = config[binkey]
    log.debug(logid + str([str(mb), str(me)]))
    return mb, me


@check_run
def sample_from_path(path):
    ret = str(os.path.join(os.path.split(str(path))[-1]))
    return ret


@check_run
def runstate_from_sample(sample, config):
    logid = scriptname + ".Params_runstate_from_sample: "
    ret = list()
    for s in sample:
        if len(s.split(os.sep)) < 2:
            s = samplecond(s, config)
        if len(mu.get_from_dict(config["SETTINGS"], s.split(os.sep))) < 1:
            s = os.path.dirname(s)
            n = s.split(os.sep)[-1]
        log.debug(logid + "SAMPLE: " + s)
        try:
            c = mu.get_from_dict(config["SETTINGS"], s.split(os.sep))[0]
        except:
            c = None
        log.debug(logid + "SETTINGS: " + str(c))
        if mu.dict_inst(c):
            if not c.get("SAMPLES"):
                for k, v in c.items():
                    log.debug(
                        logid + "k: " + str(k) + ", v: " + str(v) + " c: " + str(c)
                    )
                    if mu.dict_inst(v) and v.get("SAMPLES"):
                        if k not in ret:
                            ret.append(k)
            else:
                log.debug(logid + "c: " + str(c))
                ret.extend(s.split(os.sep))
        else:
            log.debug(logid + "c: " + str(c) + ", k: " + str(s.split(os.sep)[-1]))
            k = s.split(os.sep)[-1]
            ret.append(k)
    log.debug(logid + "RETURN: " + str(ret))
    return ret


@check_run
def samplecond(
    sample, config
):  # takes list of sample names (including .fastq.gz) and returns a list with their conditions as directory path without fastq.gz ( ["condition/of/sample", ... ])
    logid = scriptname + ".Params_samplecond: "
    ret = list()
    for s in sample:
        log.debug(logid + str(s))
        s = s.replace(".fastq.gz", "")
        check = s.split(os.sep)
        checkdir = check[:-1]
        sname = check[-1]
        tmplist = checkdir
        log.debug(logid + "CHECK: " + str(checkdir))
        for r in runstate_from_sample([s], config):
            if r not in tmplist:
                tmp = check[:-1]
                tmp.append(r)
                if sname in mu.get_from_dict(config["SETTINGS"], tmp)[0].get("SAMPLES"):
                    tmplist.append(r)
        log.debug(logid + "TMPLIST: " + str(tmplist))
        paired = checkpaired([s], config)
        if any(p in paired for p in ["paired", "singlecell"]):
            s = re.sub(r"_[r|R|][1|2]\.", "", s)
        r = os.sep.join(tmplist)
        if r not in s:
            ret.append(os.sep.join([r, os.path.basename(s)]))
        else:
            ret.append(os.sep.join([os.path.dirname(s), os.path.basename(s)]))
    log.debug(logid + "RETURN: " + str(ret))
    return ret


@check_run
def conditiononly(sample, config):
    logid = scriptname + ".Params_conditiononly: "
    ret = list()
    check = sample.split(os.sep)
    checkdir = check[:-1]
    sname = check[-1]
    ret.extend(checkdir)
    log.debug(logid + "CHECK: " + str(checkdir))
    for r in runstate_from_sample([sample], config):
        log.debug(logid + "runstate " + str(r))
        if r not in ret:
            tmp = list()
            tmp.extend(
                ret
            )  # this will take only the first occurence of sample in settings, should anyways never happen to have the same sample in different subsettings with differing pairedness
            tmp.append(r)
            log.debug(logid + "tmp: " + str(tmp))
            if len(
                mu.get_from_dict(config["SETTINGS"], tmp)
            ) > 0 and sname in mu.get_from_dict(config["SETTINGS"], tmp)[0].get(
                "SAMPLES"
            ):
                ret.append(r)
    log.debug(logid + "ret: " + str(ret))
    return ret


@check_run
def checkpaired(sample, config):
    logid = scriptname + ".Params_checkpaired: "
    paired = ""
    for s in sample:
        log.debug(logid + "SAMPLE: " + str(s))
        check = conditiononly(s, config)
        log.debug(logid + "CHECK: " + str(check))
        p = mu.sub_dict(config["SETTINGS"], check)
        if p:
            paired = p.get("SEQUENCING")
            paired = paired.split(",")[0] if "," in paired else paired
        else:
            return None
        # Per sample paired, not implemented yet
        # pairedlist = p.get('SEQUENCING')
        # samplelist = p.get('SAMPLES')
        # x = samplelist.index(s.split(os.sep)[-1])
        # paired = pairedlist[x]
    log.debug(logid + "SEQUENCING: " + str(paired))
    return paired


@check_run
def checkpaired_rep(sample, config):
    logid = scriptname + ".Params_checkpaired_rep: "
    log.debug(logid + "SAMPLE: " + str(sample))
    ret = list()
    for s in sample:
        check = conditiononly(s, config)
        p = mu.sub_dict(config["SETTINGS"], check)
        paired = p.get("SEQUENCING")
        # Per sample paired, not implemented yet
        # pairedlist = p.get('SEQUENCING')
        # samplelist = p.get('SAMPLES')
        # x = samplelist.index(s.split(os.sep)[-1])
        # paired = pairedlist[x]
        ret.append(str(paired).replace(",", "_"))
    log.debug(logid + "PAIRED: " + str(ret))
    return str.join(",", ret)


@check_run
def checkstranded(sample, config):
    logid = scriptname + ".Params_checkstranded: "
    ret = list()
    stranded = ""
    for s in sample:
        check = conditiononly(s, config)
        log.debug(logid + "check: " + str(check))
        p = mu.sub_dict(config["SETTINGS"], check)
        log.debug(logid + "P: " + str(p))
        paired = p.get("SEQUENCING")
        # Per sample paired, not implemented yet
        # pairedlist = p.get('SEQUENCING')
        # samplelist = p.get('SAMPLES')
        # x = samplelist.index(s.split(os.sep)[-1])
        # paired = pairedlist[x]
        stranded = (
            paired.split(",")[1]
            if len(paired.split(",")) > 1 and paired.split(",")[1] != "unstranded"
            else ""
        )
    log.debug(logid + "STRANDEDNESS: " + str(stranded))
    return stranded


@check_run
def set_pairing(samples, config):
    logid = scriptname + ".Params_set_pairing: "
    ret = list()
    cond = conditiononly(samples[0], config)
    pconf = mu.sub_dict(config["PEAKS"], cond)
    log.debug(logid + "SAMPLES: " + str(samples))
    pairlist = pconf.get("COMPARABLE", config["PEAKS"].get("COMPARABLE"))
    log.debug(logid + "PAIRLIST: " + str(pairlist))
    if pairlist:
        for k, v in pairlist.items():
            for x in samples:
                if str(k) == str(os.path.basename(x)):
                    ret.extend(samplecond([x], config))
    else:
        return samples
    log.debug(logid + "return: " + str(ret))
    return ret


@check_run
def get_pairing(sample, stype, config, samples, scombo="", mode="smk"):
    logid = scriptname + ".Params_get_pairing: "
    cond = conditiononly(sample, config)
    pconf = mu.sub_dict(config["PEAKS"], cond)
    pairlist = pconf.get("COMPARABLE", config["PEAKS"].get("COMPARABLE"))
    matching = ""
    log.debug(
        logid
        + "PAIRLIST: "
        + str(pairlist)
        + " SAMPLE: "
        + str(sample)
        + " SAMPLES: "
        + str(samples)
        + " Combo: "
        + str(scombo)
    )
    if pairlist:
        for k, v in pairlist.items():
            if str(k) == str(os.path.basename(sample)):
                for x in samples:
                    if str(v) == str(os.path.basename(x)) and x != sample:
                        log.debug(logid + "Match found: " + str(v) + " : " + str(x))
                        try:
                            matching = samplecond([x], config)[0].replace("MAPPED/", "")
                        except:
                            matching = x
                        log.info(logid + "PAIRINGS: " + sample + ": " + str(matching))
        if not matching or matching == "":
            log.error(
                logid
                + f"COMPARABLE set in config but no fitting pair could be found for sample {sample} in {pairlist}. Please check config."
            )
        else:
            if mode == "smk":
                retstr = (
                    f"-c {os.sep.join(['MAPPED', scombo, matching])}_mapped_{stype}.bam"
                )
            else:
                retstr = f"{os.sep.join(['MAPPED', scombo, sample])}_mapped_{stype}.bam, {os.sep.join(['MAPPED', scombo, matching])}_mapped_{stype}.bam"

            log.debug(logid + retstr)
            return retstr
    else:
        log.warning(logid + "No matching sample found")
        return ""


@check_run
def post_checkpaired(sample, config):
    logid = scriptname + ".Params_checkpaired: "
    ret = list()
    paired = ""
    for s in sample:
        log.debug(logid + "SAMPLE: " + str(sample))
        check = conditiononly(sample, config)
        p = mu.sub_dict(config["SETTINGS"], check)
        log.debug(logid + "P: " + str(p))
        paired = p.get("SEQUENCING").split(",")[0]
    log.debug(logid + "PAIRED: " + str(paired))
    return paired


@check_run
def check_IP(sample, config):
    logid = scriptname + ".Params_check_IP: "
    ret = list()
    ip = ""
    for s in sample:
        log.debug(logid + "SAMPLE: " + str(s))
        check = os.path.dirname(s).split(os.sep)
        tmplist = check
        for r in runstate_from_sample([s], config):
            if r not in tmplist:
                tmplist.extend(r)
        log.debug(logid + "TMP: " + str(tmplist))
        check = mu.get_from_dict(config["PEAKS"], tmplist)[0]
        log.debug(logid + "CHECK: " + str(check))
        if "IP" in check:
            ip = check["IP"]
        else:
            log.debug(logid + "Key IP not found in config")
    log.debug(logid + "IP is: " + str(ip))
    return str(ip)


@check_run
def check_tool_params(sample, runstate, config, subconf, idx):
    try:
        par = tool_params(sample, runstate, config, subconf)["OPTIONS"][idx]
        if par != "":
            return par
        elif subconf == "MAPPING":
            return "std"
        else:
            return ""
    except:
        if subconf == "MAPPING":
            return "std"
        else:
            return ""


@check_run
def comparable_as_string(config, subwork):
    logid = scriptname + ".comparable_as_string: "
    log.debug(logid + "this is the config: " + str(config))
    check = config[subwork].get("COMPARABLE")
    if check:
        log.debug(logid + "determine comparables in " + subwork)
        complist = []
        compdict = config[subwork]["COMPARABLE"]
        for contrast in compdict:
            As = ""
            Bs = ""
            for condition in compdict[contrast][0]:
                As = (As + "+" + condition).strip("+")
            for condition in compdict[contrast][1]:
                Bs = (Bs + "+" + condition).strip("+")
            complist.append(f"{contrast}:{As}-vs-{Bs}")
        compstr = ",".join(complist)
        return compstr
    else:
        log.warning(
            logid + "no comparables found in " + subwork + ". Compare All vs. All."
        )
        groups_by_condition = list(mu.yield_from_dict("GROUPS", config))
        flattened = sorted(
            set(val for sublist in groups_by_condition for val in sublist)
        )
        combined = list(set(itertools.combinations(flattened, 2)))
        complist = []
        for key, value in combined:
            complist.append(f"{key}vs{value}:{key}-vs-{value}")
        compstr = ",".join(complist)
        return compstr


@check_run
def get_combo_name(combinations):
    logid = scriptname + ".Params_get_combo_name: "
    combname = mu.NestedDefaultDict()

    for condition in combinations:
        log.debug(logid + "CONDITION: " + str(condition))
        combname[condition]["envs"] = list()
        combname[condition]["works"] = list()
        combos = combinations[condition]

        for combi in combos:
            envs = list()
            works = list()
            for step in combi:
                for work, env in step.items():
                    envs.append(env)
                    works.append(work)
            combname[condition]["envs"].append(str.join("-", envs))
            combname[condition]["works"].append(str.join("-", works))

    log.debug(logid + "ComboName: " + str(combname))
    return combname


#
# Params.py ends here
